#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Functions for testing the obspy.io.nordic functions
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA @UnusedWildImport

import inspect
import io
import os
import unittest
import warnings
from itertools import cycle
import numpy as np

from obspy import read_events, Catalog, UTCDateTime, read
from obspy.core.event import (
    Pick, WaveformStreamID, Arrival, Amplitude, Event, Origin, Magnitude,
    OriginQuality, EventDescription, CreationInfo, OriginUncertainty,
    ConfidenceEllipsoid, QuantityError, FocalMechanism, MomentTensor,
    NodalPlane, NodalPlanes, ResourceIdentifier, Tensor)
from obspy.core.util.base import NamedTemporaryFile
from obspy.core.util.misc import TemporaryWorkingDirectory
from obspy.core.util.testing import ImageComparison

from obspy.io.nordic import NordicParsingError
from obspy.io.nordic.core import (
    _is_sfile, read_spectral_info, read_nordic, readwavename, blanksfile,
    _write_nordic, nordpick, readheader, _readheader, write_select)
from obspy.io.nordic.utils import (
    _int_conv, _float_conv, _str_conv, _nortoevmag, _evmagtonor,
    _get_line_tags)
from obspy.io.nordic.ellipse import Ellipse


class TestNordicMethods(unittest.TestCase):
    """
    Test suite for nordic io operations.
    """
    def setUp(self):
        self.path = os.path.dirname(os.path.abspath(inspect.getfile(
            inspect.currentframe())))
        self.testing_path = os.path.join(self.path, "data")

    def test_read_write(self):
        """
        Function to test the read and write capabilities of sfile_util.
        """
        # Set-up a test event
        test_event = full_test_event()
        # Sort the magnitudes - they are sorted on writing and we need to check
        # like-for-like
        test_event.magnitudes.sort(key=lambda obj: obj['mag'], reverse=True)
        # Add the event to a catalogue which can be used for QuakeML testing
        test_cat = Catalog()
        test_cat += test_event
        # Check the read-write s-file functionality
        with TemporaryWorkingDirectory():
            with warnings.catch_warnings():
                # Evaluation mode mapping warning
                warnings.simplefilter('ignore', UserWarning)
                sfile = _write_nordic(
                    test_cat[0], filename=None, userid='TEST', evtype='L',
                    outdir='.', wavefiles='test', explosion=True,
                    overwrite=True)
            self.assertEqual(readwavename(sfile), ['test'])
            read_cat = Catalog()
            # raises "UserWarning: AIN in header, currently unsupported"
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                read_cat += read_nordic(sfile)
        read_ev = read_cat[0]
        test_ev = test_cat[0]
        for read_pick, test_pick in zip(read_ev.picks, test_ev.picks):
            self.assertEqual(read_pick.time, test_pick.time)
            self.assertEqual(read_pick.backazimuth, test_pick.backazimuth)
            self.assertEqual(read_pick.onset, test_pick.onset)
            self.assertEqual(read_pick.phase_hint, test_pick.phase_hint)
            if test_pick.polarity == "undecidable":
                self.assertIsNone(read_pick.polarity)
            elif read_pick.polarity == "undecidable":
                self.assertIsNone(test_pick.polarity)
            else:
                self.assertEqual(read_pick.polarity, test_pick.polarity)
            self.assertEqual(read_pick.waveform_id.station_code,
                             test_pick.waveform_id.station_code)
            self.assertEqual(read_pick.waveform_id.channel_code[-1],
                             test_pick.waveform_id.channel_code[-1])
        # assert read_ev.origins[0].resource_id ==\
        #     test_ev.origins[0].resource_id
        self.assertEqual(read_ev.origins[0].time,
                         test_ev.origins[0].time)
        # Note that time_residual_RMS is not a quakeML format
        self.assertEqual(read_ev.origins[0].longitude,
                         test_ev.origins[0].longitude)
        self.assertEqual(read_ev.origins[0].latitude,
                         test_ev.origins[0].latitude)
        self.assertEqual(read_ev.origins[0].depth,
                         test_ev.origins[0].depth)
        self.assertEqual(read_ev.magnitudes[0].mag,
                         test_ev.magnitudes[0].mag)
        self.assertEqual(read_ev.magnitudes[1].mag,
                         test_ev.magnitudes[1].mag)
        self.assertEqual(read_ev.magnitudes[2].mag,
                         test_ev.magnitudes[2].mag)
        self.assertEqual(read_ev.magnitudes[0].creation_info,
                         test_ev.magnitudes[0].creation_info)
        self.assertEqual(read_ev.magnitudes[1].creation_info,
                         test_ev.magnitudes[1].creation_info)
        self.assertEqual(read_ev.magnitudes[2].creation_info,
                         test_ev.magnitudes[2].creation_info)
        self.assertEqual(read_ev.magnitudes[0].magnitude_type,
                         test_ev.magnitudes[0].magnitude_type)
        self.assertEqual(read_ev.magnitudes[1].magnitude_type,
                         test_ev.magnitudes[1].magnitude_type)
        self.assertEqual(read_ev.magnitudes[2].magnitude_type,
                         test_ev.magnitudes[2].magnitude_type)
        self.assertEqual(read_ev.event_descriptions,
                         test_ev.event_descriptions)
        # assert read_ev.amplitudes[0].resource_id ==\
        #     test_ev.amplitudes[0].resource_id
        self.assertEqual(read_ev.amplitudes[0].period,
                         test_ev.amplitudes[0].period)
        self.assertEqual(read_ev.amplitudes[0].snr,
                         test_ev.amplitudes[0].snr)
        self.assertEqual(read_ev.amplitudes[2].period,
                         test_ev.amplitudes[2].period)
        self.assertEqual(read_ev.amplitudes[2].snr,
                         test_ev.amplitudes[2].snr)
        # Check coda magnitude pick
        # Resource ids get overwritten because you can't have two the same in
        # memory
        # self.assertEqual(read_ev.amplitudes[1].resource_id,
        #                  test_ev.amplitudes[1].resource_id)
        self.assertEqual(read_ev.amplitudes[1].type,
                         test_ev.amplitudes[1].type)
        self.assertEqual(read_ev.amplitudes[1].unit,
                         test_ev.amplitudes[1].unit)
        self.assertEqual(read_ev.amplitudes[1].generic_amplitude,
                         test_ev.amplitudes[1].generic_amplitude)
        # Resource ids get overwritten because you can't have two the same in
        # memory
        # self.assertEqual(read_ev.amplitudes[1].pick_id,
        #                  test_ev.amplitudes[1].pick_id)
        self.assertEqual(read_ev.amplitudes[1].waveform_id.station_code,
                         test_ev.amplitudes[1].waveform_id.station_code)
        self.assertEqual(read_ev.amplitudes[1].waveform_id.channel_code,
                         test_ev.amplitudes[1].waveform_id.channel_code[0] +
                         test_ev.amplitudes[1].waveform_id.channel_code[-1])
        self.assertEqual(read_ev.amplitudes[1].magnitude_hint,
                         test_ev.amplitudes[1].magnitude_hint)
        # snr is not supported in s-file
        # self.assertEqual(read_ev.amplitudes[1].snr,
        #                  test_ev.amplitudes[1].snr)
        self.assertEqual(read_ev.amplitudes[1].category,
                         test_ev.amplitudes[1].category)

    def test_fail_writing(self):
        """
        Test a deliberate fail.
        """
        test_event = full_test_event()
        # Add the event to a catalogue which can be used for QuakeML testing
        test_cat = Catalog()
        test_cat += test_event
        test_ev = test_cat[0]
        test_cat.append(full_test_event())
        with self.assertRaises(NordicParsingError):
            # Raises error due to multiple events in catalog
            with warnings.catch_warnings():
                # Evaluation mode mapping warning
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(test_cat, filename=None, userid='TEST',
                              evtype='L', outdir='.',
                              wavefiles='test', explosion=True,
                              overwrite=True)
        with self.assertRaises(NordicParsingError):
            # Raises error due to too long userid
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(test_ev, filename=None, userid='TESTICLE',
                              evtype='L', outdir='.',
                              wavefiles='test', explosion=True,
                              overwrite=True)
        with self.assertRaises(NordicParsingError):
            # Raises error due to unrecognised event type
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(test_ev, filename=None, userid='TEST',
                              evtype='U', outdir='.',
                              wavefiles='test', explosion=True,
                              overwrite=True)
        with self.assertRaises(NordicParsingError):
            # Raises error due to no output directory
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(test_ev, filename=None, userid='TEST',
                              evtype='L', outdir='albatross',
                              wavefiles='test', explosion=True,
                              overwrite=True)
        invalid_origin = test_ev.copy()

        invalid_origin.origins = []
        with self.assertRaises(NordicParsingError):
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(invalid_origin, filename=None, userid='TEST',
                              evtype='L', outdir='.', wavefiles='test',
                              explosion=True, overwrite=True)
        invalid_origin = test_ev.copy()
        invalid_origin.origins[0].time = None
        with self.assertRaises(NordicParsingError):
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(invalid_origin, filename=None, userid='TEST',
                              evtype='L', outdir='.', wavefiles='test',
                              explosion=True, overwrite=True)
        # Write a near empty origin
        valid_origin = test_ev.copy()
        valid_origin.origins[0].latitude = None
        valid_origin.origins[0].longitude = None
        valid_origin.origins[0].depth = None
        with NamedTemporaryFile() as tf:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(
                    valid_origin, filename=tf.name, userid='TEST',
                    evtype='L', outdir='.', wavefiles='test',
                    explosion=True, overwrite=True)
            self.assertTrue(os.path.isfile(tf.name))

    def test_blanksfile(self):
        st = read()
        with TemporaryWorkingDirectory():
            testing_path = 'Temporary_wavefile'
            st.write(testing_path, format='MSEED')
            sfile = blanksfile(testing_path, 'L', 'TEST', overwrite=True)
            self.assertTrue(os.path.isfile(sfile))
            os.remove(sfile)
            sfile = blanksfile(testing_path, 'L', 'TEST', overwrite=True,
                               evtime=UTCDateTime())
            self.assertTrue(os.path.isfile(sfile))
            os.remove(sfile)
            with self.assertRaises(NordicParsingError):
                # No wavefile
                blanksfile('albert', 'L', 'TEST', overwrite=True)
            with self.assertRaises(NordicParsingError):
                # USER ID too long
                blanksfile(testing_path, 'L', 'TESTICLE', overwrite=True)
            with self.assertRaises(NordicParsingError):
                # Unknown event type
                blanksfile(testing_path, 'U', 'TEST', overwrite=True)
            # Check that it breaks when writing multiple versions
            sfiles = []
            for _i in range(10):
                # raises UserWarning: Desired sfile exists, will not overwrite
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', UserWarning)
                    sfiles.append(blanksfile(testing_path, 'L', 'TEST'))
            with self.assertRaises(NordicParsingError):
                # raises UserWarning: Desired sfile exists, will not overwrite
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', UserWarning)
                    blanksfile(testing_path, 'L', 'TEST')
            for sfile in sfiles:
                self.assertTrue(os.path.isfile(sfile))

    def test_write_empty(self):
        """
        Function to check that writing a blank event works as it should.
        """
        test_event = Event()
        with self.assertRaises(NordicParsingError):
            _write_nordic(test_event, filename=None, userid='TEST', evtype='L',
                          outdir='.', wavefiles='test')
        test_event.origins.append(Origin())
        with self.assertRaises(NordicParsingError):
            _write_nordic(test_event, filename=None, userid='TEST', evtype='L',
                          outdir='.', wavefiles='test')
        test_event.origins[0].time = UTCDateTime()
        with TemporaryWorkingDirectory():
            test_sfile = _write_nordic(test_event, filename=None,
                                       userid='TEST', evtype='L', outdir='.',
                                       wavefiles='test')
            self.assertTrue(os.path.isfile(test_sfile))

    def test_read_empty_header(self):
        """
        Function to check a known issue, empty header info S-file: Bug found \
        by Dominic Evanzia.
        """
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)

            test_event = read_nordic(os.path.join(self.testing_path,
                                                  'Sfile_no_location'))[0]
        self.assertFalse(test_event.origins[0].latitude)
        self.assertFalse(test_event.origins[0].longitude)
        self.assertFalse(test_event.origins[0].depth)

    def test_read_extra_header(self):
        testing_path = os.path.join(self.testing_path, 'Sfile_extra_header')
        not_extra_header = os.path.join(self.testing_path,
                                        '01-0411-15L.S201309')
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            test_event = read_nordic(testing_path)[0]
            header_event = read_nordic(not_extra_header)[0]
        self.assertEqual(len(header_event.origins), 2)
        self.assertEqual(test_event.origins[0].time,
                         header_event.origins[0].time)
        self.assertEqual(test_event.origins[0].latitude,
                         header_event.origins[0].latitude)
        self.assertEqual(test_event.origins[0].longitude,
                         header_event.origins[0].longitude)
        self.assertEqual(test_event.origins[0].depth,
                         header_event.origins[0].depth)

    def test_header_mapping(self):
        # Raise "UserWarning: Lines of type I..."
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            head_1 = readheader(os.path.join(self.testing_path,
                                             '01-0411-15L.S201309'))
        with open(os.path.join(self.testing_path,
                               '01-0411-15L.S201309'), 'r') as f:
            # raises "UserWarning: AIN in header, currently unsupported"
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                tagged_lines = _get_line_tags(f=f)
                head_2 = _readheader(head_lines=tagged_lines['1'])
        _assert_similarity(head_1, head_2)

    def test_missing_header(self):
        # Check that a suitable error is raised
        with self.assertRaises(NordicParsingError):
            # Raises AIN warning
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                readheader(os.path.join(self.testing_path, 'Sfile_no_header'))

    def test_reading_string_io(self):
        filename = os.path.join(self.testing_path, '01-0411-15L.S201309')
        with open(filename, "rt") as fh:
            file_object = io.StringIO(fh.read())

        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)

            cat = read_events(file_object)
            file_object.close()

            ref_cat = read_events(filename)
            _assert_similarity(cat[0], ref_cat[0])

    def test_reading_bytes_io(self):
        filename = os.path.join(self.testing_path, '01-0411-15L.S201309')
        with open(filename, "rb") as fh:
            file_object = io.BytesIO(fh.read())

        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)

            cat = read_events(file_object)
            file_object.close()

            ref_cat = read_events(filename)
            _assert_similarity(cat[0], ref_cat[0])

    def test_corrupt_header(self):
        filename = os.path.join(self.testing_path, '01-0411-15L.S201309')
        f = open(filename, 'r')
        with NamedTemporaryFile(suffix='.sfile') as tmp_file:
            fout = open(tmp_file.name, 'w')
            for line in f:
                fout.write(line[0:78])
            f.close()
            fout.close()
            with self.assertRaises(NordicParsingError):
                readheader(tmp_file.name)

    def test_multi_writing(self):
        event = full_test_event()
        # Try to write the same event multiple times, but not overwrite
        sfiles = []
        with TemporaryWorkingDirectory():
            for _i in range(59):
                with warnings.catch_warnings():
                    # Evaluation mode mapping warning
                    warnings.simplefilter('ignore', UserWarning)
                    sfiles.append(_write_nordic(event=event, filename=None,
                                                overwrite=False))
            with self.assertRaises(NordicParsingError):
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', UserWarning)
                    _write_nordic(event=event, filename=None, overwrite=False)

    def test_mag_conv(self):
        """
        Check that we convert magnitudes as we should!
        """
        magnitude_map = [
            ('L', 'ML'), ('B', 'mB'), ('S', 'MS'), ('W', 'MW'), ('G', 'MbLg'),
            ('C', 'Mc'), ('s', 'Ms')]
        for magnitude in magnitude_map:
            self.assertEqual(magnitude[0], _evmagtonor(magnitude[1]))
            self.assertEqual(_nortoevmag(magnitude[0]), magnitude[1])

    def test_str_conv(self):
        """
        Test the simple string conversions.
        """
        self.assertEqual(_int_conv('albert'), None)
        self.assertEqual(_float_conv('albert'), None)
        self.assertEqual(_str_conv('albert'), 'albert')
        self.assertEqual(_int_conv('1'), 1)
        self.assertEqual(_float_conv('1'), 1.0)
        self.assertEqual(_str_conv(1), '1')
        self.assertEqual(_int_conv('1.0256'), None)
        self.assertEqual(_float_conv('1.0256'), 1.0256)
        self.assertEqual(_str_conv(1.0256), '1.0256')

    def test_read_wavename(self):
        testing_path = os.path.join(self.testing_path, '01-0411-15L.S201309')
        wavefiles = readwavename(testing_path)
        self.assertEqual(len(wavefiles), 1)
        # Check that read_nordic reads wavname when return_wavnames=True
        cat, wavefiles = read_nordic(testing_path, return_wavnames=True)
        self.assertEqual(wavefiles, [['2013-09-01-0410-35.DFDPC_024_00']])
        # Test that full paths are handled
        test_event = full_test_event()
        # Add the event to a catalogue which can be used for QuakeML testing
        test_cat = Catalog()
        test_cat += test_event
        # Check the read-write s-file functionality
        with TemporaryWorkingDirectory():
            with warnings.catch_warnings():
                # Evaluation mode mapping warning
                warnings.simplefilter('ignore', UserWarning)
                sfile = _write_nordic(
                    test_cat[0], filename=None, userid='TEST', evtype='L',
                    outdir='.', wavefiles=['walrus/test'], explosion=True,
                    overwrite=True)
            self.assertEqual(readwavename(sfile), ['test'])
        # Check that multiple wavefiles are read properly
        with TemporaryWorkingDirectory():
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                sfile = _write_nordic(
                    test_cat[0], filename=None, userid='TEST', evtype='L',
                    outdir='.', wavefiles=['walrus/test', 'albert'],
                    explosion=True, overwrite=True)
            self.assertEqual(readwavename(sfile), ['test', 'albert'])

    def test_read_event(self):
        """
        Test the wrapper.
        """
        testing_path = os.path.join(self.testing_path, '01-0411-15L.S201309')
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            event = read_nordic(testing_path)[0]
        self.assertEqual(len(event.origins), 2)
        self.assertEqual(len(event.picks), 17)

    def test_read_latin1(self):
        """
        Check that we can read dos formatted, latin1 encoded files.
        """
        with warnings.catch_warnings():
            # Lots of unsupported line warnings
            warnings.simplefilter('ignore', UserWarning)
            dos_file = os.path.join(self.testing_path, 'dos-file.sfile')
            self.assertTrue(_is_sfile(dos_file))
            event = readheader(dos_file)
            self.assertEqual(event.origins[0].latitude, 60.328)
            cat = read_events(dos_file)
            self.assertEqual(cat[0].origins[0].latitude, 60.328)
            wavefiles = readwavename(dos_file)
            self.assertEqual(wavefiles[0], "90121311.0851W41")
            spectral_info = read_spectral_info(dos_file)
            self.assertEqual(len(spectral_info.keys()), 10)
            self.assertEqual(spectral_info[('AVERAGE', '')]['stress_drop'],
                             27.7)
            with self.assertRaises(UnicodeDecodeError):
                readheader(dos_file, 'ASCII')

    def test_read_many_events(self):
        testing_path = os.path.join(self.testing_path, 'select.out')
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            catalog = read_nordic(testing_path)
        self.assertEqual(len(catalog), 50)

    def test_write_select(self):
        cat = read_events()
        with NamedTemporaryFile(suffix='.out') as tf:
            # raises "UserWarning: mb is not convertible"
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                write_select(cat, filename=tf.name)
            self.assertTrue(_is_sfile(tf.name))
            with warnings.catch_warnings():
                # Type I warning
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
        for event_1, event_2 in zip(cat, cat_back):
            _assert_similarity(event_1=event_1, event_2=event_2)

    def test_write_plugin(self):
        cat = read_events()
        cat.append(full_test_event())
        with NamedTemporaryFile(suffix='.out') as tf:
            # raises UserWarning: mb is not convertible
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat.write(tf.name, format='nordic')
            # raises "UserWarning: AIN in header, currently unsupported"
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
            for event_1, event_2 in zip(cat, cat_back):
                _assert_similarity(event_1=event_1, event_2=event_2)

    def test_more_than_three_mags(self):
        cat = Catalog()
        cat += full_test_event()
        cat[0].magnitudes.append(Magnitude(
            mag=0.9, magnitude_type='MS', creation_info=CreationInfo('TES'),
            origin_id=cat[0].origins[0].resource_id))
        with NamedTemporaryFile(suffix='.out') as tf:
            # raises UserWarning: mb is not convertible
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat.write(tf.name, format='nordic')
            # raises "UserWarning: AIN in header, currently unsupported"
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
            for event_1, event_2 in zip(cat, cat_back):
                self.assertTrue(
                    len(event_1.magnitudes) == len(event_2.magnitudes))
                _assert_similarity(event_1, event_2)

    def test_inaccurate_picks(self):
        testing_path = os.path.join(self.testing_path, 'bad_picks.sfile')
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_nordic(testing_path)
        pick_string = nordpick(cat[0])
        for pick in pick_string:
            self.assertEqual(len(pick), 80)

    def test_round_len(self):
        testing_path = os.path.join(self.testing_path, 'round_len_undef.sfile')
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            event = read_nordic(testing_path)[0]
        pick_string = nordpick(event)
        for pick in pick_string:
            self.assertEqual(len(pick), 80)

    def test_read_moment(self):
        """
        Test the reading of seismic moment from the s-file.
        """
        testing_path = os.path.join(self.testing_path, 'automag.out')
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            event = read_nordic(testing_path)[0]
        mag = [m for m in event.magnitudes if m.magnitude_type == 'MW']
        self.assertEqual(len(mag), 1)
        self.assertEqual(mag[0].mag, 0.7)

    def test_read_moment_info(self):
        """
        Test reading the info from spectral analysis.
        """
        testing_path = os.path.join(self.testing_path, 'automag.out')
        with warnings.catch_warnings():
            # Userwarning, type I
            warnings.simplefilter('ignore', UserWarning)
            spec_inf = read_spectral_info(testing_path)
        self.assertEqual(len(spec_inf), 5)
        # This should actually test that what we are reading in is correct.
        average = spec_inf[('AVERAGE', '')]
        check_av = {u'channel': '', u'corner_freq': 5.97, u'decay': 0.0,
                    u'moment': 12589254117.941662, u'moment_mag': 0.7,
                    u'source_radius': 0.231,
                    u'spectral_level': 0.3981071705534972,
                    u'station': 'AVERAGE', u'stress_drop': 0.006,
                    u'window_length': 1.6}
        for key in average.keys():
            if isinstance(average.get(key), str):
                self.assertEqual(average.get(key), check_av.get(key))
            else:
                self.assertEqual(round(average.get(key), 4),
                                 round(check_av.get(key), 4))

    def test_is_sfile(self):
        sfiles = ['01-0411-15L.S201309', 'automag.out', 'bad_picks.sfile',
                  'round_len_undef.sfile', 'Sfile_extra_header',
                  'Sfile_no_location']
        for sfile in sfiles:
            self.assertTrue(_is_sfile(os.path.join(self.testing_path, sfile)))
        self.assertFalse(
            _is_sfile(os.path.join(self.testing_path, 'Sfile_no_header')))
        self.assertFalse(_is_sfile(os.path.join(
            self.path, '..', '..', 'nlloc', 'tests', 'data', 'nlloc.hyp')))

    def test_read_picks_across_day_end(self):
        testing_path = os.path.join(self.testing_path, 'sfile_over_day')
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            event = read_nordic(testing_path)[0]
        pick_times = [pick.time for pick in event.picks]
        for pick in event.picks:
            # Pick should come after origin time
            self.assertGreater(pick.time, event.origins[0].time)
            # All picks in this event are within 60s of origin time
            self.assertLessEqual((pick.time - event.origins[0].time), 60)
        # Make sure zero hours and 24 hour picks are handled the same.
        testing_path = os.path.join(self.testing_path, 'sfile_over_day_zeros')
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            event_2 = read_nordic(testing_path)[0]
        for pick in event_2.picks:
            # Pick should come after origin time
            self.assertGreater(pick.time, event_2.origins[0].time)
            # All picks in this event are within 60s of origin time
            self.assertLessEqual((pick.time - event_2.origins[0].time), 60)
            # Each pick should be the same as one pick in the previous event
            self.assertTrue(pick.time in pick_times)
        self.assertEqual(event_2.origins[0].time, event.origins[0].time)

    def test_distance_conversion(self):
        """
        Check that distances are converted properly.
        """
        testing_path = os.path.join(self.testing_path, '01-0411-15L.S201309')
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(testing_path)
        event = cat[0]
        self.assertAlmostEqual(
            sorted(event.origins[0].arrivals,
                   key=lambda x: x.distance)[0].distance, 0.035972864236749225)
        pick_strings = nordpick(event)
        self.assertEqual(
            int([p for p in pick_strings if p.split()[0] == 'GCSZ' and
                 p.split()[1] == 'SZ'][0].split()[-1]), 304)
        self.assertEqual(
            int([p for p in pick_strings if p.split()[0] == 'WZ11' and
                 p.split()[1] == 'HZ'][0].split()[-1]), 30)

    def test_large_negative_longitude(self):
        event = full_test_event()
        event.origins[0].longitude = -120
        with NamedTemporaryFile(suffix=".out") as tf:
            with warnings.catch_warnings():
                # Evaluation mode mapping warning
                warnings.simplefilter('ignore', UserWarning)
                event.write(tf.name, format="NORDIC")
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                event_back = read_events(tf.name)
        _assert_similarity(event, event_back[0])

    def test_write_preferred_origin(self):
        event = full_test_event()
        preferred_origin = Origin(
            time=UTCDateTime("2012-03-26") + 2.2, latitude=47.0,
            longitude=35.0, depth=18000)
        event.origins.append(preferred_origin)
        event.preferred_origin_id = preferred_origin.resource_id
        with NamedTemporaryFile(suffix=".out") as tf:
            with warnings.catch_warnings():
                # Evaluation mode mapping warning
                warnings.simplefilter('ignore', UserWarning)
                event.write(tf.name, format="NORDIC")
            with warnings.catch_warnings():
                # Type I warning
                warnings.simplefilter('ignore', UserWarning)
                event_back = read_events(tf.name)
        self.assertEqual(preferred_origin.latitude,
                         event_back[0].origins[0].latitude)
        self.assertEqual(preferred_origin.longitude,
                         event_back[0].origins[0].longitude)
        self.assertEqual(preferred_origin.depth,
                         event_back[0].origins[0].depth)
        self.assertEqual(preferred_origin.time,
                         event_back[0].origins[0].time)

    def test_read_high_precision_pick(self):
        """
        Nordic supports writing to milliseconds in high-precision mode,
        obspy < 1.2.0 did not properly read this, see #2348.
        """
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(
                os.path.join(self.testing_path, "sfile_high_precision_picks"))
        event = cat[0]
        pick_times = {
            "LSd1": UTCDateTime(2010, 11, 26, 1, 28, 46.859),
            "LSd3": UTCDateTime(2010, 11, 26, 1, 28, 48.132),
            "LSd2": UTCDateTime(2010, 11, 26, 1, 28, 48.183),
            "LSd4": UTCDateTime(2010, 11, 26, 1, 28, 49.744)}
        for key, value in pick_times.items():
            pick = [p for p in event.picks
                    if p.waveform_id.station_code == key]
            self.assertEqual(len(pick), 1)
            self.assertEqual(pick[0].time, value)

    def test_high_precision_read_write(self):
        """ Test that high-precision writing works. """
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(
                os.path.join(self.testing_path, "sfile_high_precision_picks"))
        event = cat[0]
        pick_times = {
            "LSd1": UTCDateTime(2010, 11, 26, 1, 28, 46.859),
            "LSd3": UTCDateTime(2010, 11, 26, 1, 28, 48.132),
            "LSd2": UTCDateTime(2010, 11, 26, 1, 28, 48.183),
            "LSd4": UTCDateTime(2010, 11, 26, 1, 28, 49.744)}
        for key, value in pick_times.items():
            pick = [p for p in event.picks
                    if p.waveform_id.station_code == key]
            self.assertEqual(len(pick), 1)
            self.assertEqual(pick[0].time, value)
        with NamedTemporaryFile(suffix=".out") as tf:
            write_select(cat, filename=tf.name)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
        self.assertEqual(len(cat_back), 1)
        for key, value in pick_times.items():
            pick = [p for p in cat_back[0].picks
                    if p.waveform_id.station_code == key]
            self.assertEqual(len(pick), 1)
            self.assertEqual(pick[0].time, value)
        # Check that writing to standard accuracy just gives a rounded version
        with NamedTemporaryFile(suffix=".out") as tf:
            cat.write(format="NORDIC", filename=tf.name, high_accuracy=False)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
        self.assertEqual(len(cat_back), 1)
        for key, value in pick_times.items():
            pick = [p for p in cat_back[0].picks
                    if p.waveform_id.station_code == key]
            self.assertEqual(len(pick), 1)
            rounded_pick_time = UTCDateTime(
                value.year, value.month, value.day, value.hour, value.minute)
            rounded_pick_time += round(
                value.second + (value.microsecond / 1e6), 2)
            self.assertEqual(pick[0].time, rounded_pick_time)

    def test_long_phase_name(self):
        """ Nordic format supports 8 char phase names, sometimes. """
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(
                os.path.join(self.testing_path, "sfile_long_phase"))
        # This file has one event with one pick
        pick = cat[0].picks[0]
        arrival = cat[0].origins[0].arrivals[0]
        self.assertEqual(pick.phase_hint, "PKiKP")
        self.assertEqual(arrival.time_weight, 1)
        with NamedTemporaryFile(suffix=".out") as tf:
            write_select(cat, filename=tf.name)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
        pick = cat_back[0].picks[0]
        arrival = cat_back[0].origins[0].arrivals[0]
        self.assertEqual(pick.phase_hint, "PKiKP")
        self.assertEqual(arrival.time_weight, 1)

    def test_read_write_over_day(self):
        """
        Nordic picks are relative to origin time - check that this works
        over day boundaries.
        """
        event = full_test_event()
        event.origins[0].time -= 3600
        self.assertGreater(
            event.picks[0].time.date, event.origins[0].time.date)
        with NamedTemporaryFile(suffix=".out") as tf:
            with warnings.catch_warnings():
                # Evaluation mode mapping warning
                warnings.simplefilter('ignore', UserWarning)
                write_select(Catalog([event]), filename=tf.name)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                event_back = read_events(tf.name)[0]
        _assert_similarity(event, event_back)

    def test_seconds_overflow(self):
        """
        #2348 indicates that SEISAN sometimes overflows seconds into column 29.
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(
                os.path.join(self.testing_path, "sfile_seconds_overflow"))
        event = cat[0]
        pick_times = {
            "LSb2": UTCDateTime(2009, 7, 2, 6, 49) + 100.24}
        for key, value in pick_times.items():
            pick = [p for p in event.picks
                    if p.waveform_id.station_code == key]
            self.assertEqual(len(pick), 1)
            self.assertEqual(pick[0].time, value)
        with NamedTemporaryFile(suffix=".out") as tf:
            write_select(cat, filename=tf.name)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
        self.assertEqual(len(cat_back), 1)
        for key, value in pick_times.items():
            pick = [p for p in cat_back[0].picks
                    if p.waveform_id.station_code == key]
            self.assertEqual(len(pick), 1)
            self.assertEqual(pick[0].time, value)

    def test_read_high_accuracy(self):
        """
        Verify that high-accuracy locations are read, if present
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(
                os.path.join(self.testing_path, "sfile_highaccuracy"))
        event = cat[0]
        event_time = event.origins[0].time
        event_lat = event.origins[0].latitude
        event_lon = event.origins[0].longitude
        event_depth = event.origins[0].depth
        event_rms = event.origins[0].quality.standard_error
        self.assertEqual(event_time,
                         UTCDateTime(2015, 4, 24, 15, 25, 37) + 0.676)
        self.assertEqual(event_lat, 37.29242)
        self.assertEqual(event_lon, -32.26983)
        self.assertEqual(event_depth, 1969.)
        self.assertEqual(event_rms, 0.051)

    def test_ellipse_from__to_uncerts(self):
        """
        Verify ellipse is properly calculated and inverted using uncertainties

        tests Ellipse.from_uncerts and Ellipse.to_uncerts()
        """
        center = (20, 30)
        # First try simple cases without correlation
        x_errs = (0.5, 1.33, 1.0)
        y_errs = (1.33, 0.5, 1.0)
        for c_xy in [0, 0.2, 0.4, 0.6]:
            for (x_err, y_err) in zip(x_errs, y_errs):
                ell = Ellipse.from_uncerts(x_err, y_err, c_xy, center)
                (x_err_out, y_err_out, c_xy_out, center_out) = ell.to_uncerts()
                self.assertAlmostEqual(x_err, x_err_out)
                self.assertAlmostEqual(y_err, y_err_out)
                self.assertAlmostEqual(c_xy, c_xy_out)
                self.assertAlmostEqual(center, center_out)
        # Now a specific case with a finite covariance
        x_err = 0.5
        y_err = 1.1
        c_xy = -0.2149
        # Calculate ellipse
        ell = Ellipse.from_uncerts(x_err, y_err, c_xy, center)
        self.assertAlmostEqual(ell.a, 1.120674193646)
        self.assertAlmostEqual(ell.b, 0.451762494786)
        self.assertAlmostEqual(ell.theta, 167.9407699)
        # Calculate covariance error from ellipse
        (x_err_out, y_err_out, c_xy_out, center_out) = ell.to_uncerts()
        self.assertAlmostEqual(x_err, x_err_out)
        self.assertAlmostEqual(y_err, y_err_out)
        self.assertAlmostEqual(c_xy, c_xy_out)
        self.assertAlmostEqual(center, center_out)

    def test_ellipse_from_to_cov(self):
        """
        Verify ellipse is properly calculated and inverted using covariance

        tests Ellipse.from_uncerts and Ellipse.to_uncerts()
        """
        center = (20, 30)
        x_err = 0.5
        y_err = 1.1
        c_xy = -0.2149
        cov = [[x_err**2, c_xy], [c_xy, y_err**2]]
        # Calculate ellipse
        ell = Ellipse.from_cov(cov, center)
        self.assertAlmostEqual(ell.a, 1.120674193646)
        self.assertAlmostEqual(ell.b, 0.451762494786)
        self.assertAlmostEqual(ell.theta, 167.9407699)
        # Calculate covariance error from ellipse
        cov_out, center_out = ell.to_cov()
        self.assertAlmostEqual(cov[0][0], cov_out[0][0])
        self.assertAlmostEqual(cov[0][1], cov_out[0][1])
        self.assertAlmostEqual(cov[1][0], cov_out[1][0])
        self.assertAlmostEqual(cov[1][1], cov_out[1][1])

    def test_ellipse_from_uncerts_baz(self, debug=False):
        """
        Verify alternative ellipse creator

        tests Ellipse.from_uncerts_baz
        """
        # Now a specific case with a finite covariance
        x_err = 0.5
        y_err = 1.1
        c_xy = -0.2149
        dist = 10
        baz = 90
        viewpoint = (5, 5)
        # Calculate ellipse
        ell = Ellipse.from_uncerts_baz(x_err, y_err, c_xy,
                                       dist, baz, viewpoint)
        self.assertAlmostEqual(ell.a, 1.120674193646)
        self.assertAlmostEqual(ell.b, 0.451762494786)
        self.assertAlmostEqual(ell.theta, 167.9407699)
        self.assertAlmostEqual(ell.x, 15)
        self.assertAlmostEqual(ell.y, 5)
        baz = 180
        ell = Ellipse.from_uncerts_baz(x_err, y_err, c_xy,
                                       dist, baz, viewpoint)
        self.assertAlmostEqual(ell.x, 5)
        self.assertAlmostEqual(ell.y, -5)

    def test_ellipse_is_inside(self, debug=False):
        """
        Verify Ellipse.is_inside()
        """
        ell = Ellipse(20, 10, 90)
        self.assertIs(ell.is_inside((0, 0)), True)
        self.assertFalse(ell.is_inside((100, 100)))
        self.assertTrue(ell.is_inside((-19.9, 0)))
        self.assertTrue(ell.is_inside((19.9, 0)))
        self.assertFalse(ell.is_inside((-20.1, 0)))
        self.assertFalse(ell.is_inside((20.1, 0)))
        self.assertTrue(ell.is_inside((0, 9.9)))
        self.assertTrue(ell.is_inside((0, -9.9)))
        self.assertFalse(ell.is_inside((0, 10.1)))
        self.assertFalse(ell.is_inside((0, -10.1)))

    def test_ellipse_is_on(self, debug=False):
        """
        Verify Ellipse.is_on()
        """
        ell = Ellipse(20, 10, 90)
        self.assertFalse(ell.is_on((0, 0)))
        self.assertFalse(ell.is_on((100, 100)))
        self.assertTrue(ell.is_on((-20, 0)))
        self.assertTrue(ell.is_on((20, 0)))
        self.assertFalse(ell.is_on((-20.1, 0)))
        self.assertFalse(ell.is_on((20.1, 0)))
        self.assertTrue(ell.is_on((0, 10)))
        self.assertTrue(ell.is_on((0, -10)))
        self.assertFalse(ell.is_on((0, 10.1)))
        self.assertFalse(ell.is_on((0, -10.1)))

    def test_ellipse_subtended_angle(self, debug=False):
        """
        Verify Ellipse.subtended_angle()
        """
        ell = Ellipse(20, 10, 90)
        self.assertAlmostEqual(ell.subtended_angle((20, 0)), 180.)
        self.assertAlmostEqual(ell.subtended_angle((0, 0)), 360.)
        self.assertAlmostEqual(ell.subtended_angle((40, 0)), 32.2042275039720)
        self.assertAlmostEqual(ell.subtended_angle((0, 40)), 54.6234598480584)
        self.assertAlmostEqual(ell.subtended_angle((20, 10)), 89.9994270422)

    def test_ellipse_plot(self):
        """
        Test Ellipse.plot()

        To generate test figures, used same commands after:
        from ellipse import Ellipse
        import matplotlib.pyplot as plt
        plt.style.use('classic')
        """
        # Test single ellipse
        with ImageComparison(self.testing_path, 'plot_ellipse.png',
                             style='classic', reltol=10) as ic:
            Ellipse(20, 10, 90).plot(outfile=ic.name)
        # Test multi-ellipse figure
        with ImageComparison(self.testing_path, 'plot_ellipses.png',
                             style='classic', reltol=10) as ic:
            fig = Ellipse(20, 10, 90).plot(color='r')
            fig = Ellipse(20, 10, 45).plot(fig=fig, color='b')
            fig = Ellipse(20, 10, 0, center=(10, 10)).plot(fig=fig, color='g')
            fig = Ellipse(20, 10, -45).plot(fig=fig, outfile=ic.name)

    def test_ellipse_plot_tangents(self):
        """
        Test Ellipse.plot_tangents()
        """
        import matplotlib.pyplot as plt
        # Test single ellipse and point
        with ImageComparison(self.testing_path, 'plot_ellipse_tangents.png',
                             style='classic', reltol=10) as ic:
            Ellipse(20, 10, 90).plot_tangents((30, 30),
                                              color='b',
                                              print_angle=True,
                                              ellipse_name='Ellipse',
                                              outfile=ic.name)
        # Test multi-ellipse figure
        dist = 50
        fig = None
        try:
            prop_cycle = plt.rcParams['axes.prop_cycle']
        # prop_cycle was introduced at some point between mpl 1.x and 2.0 it
        # seems
        # XXX workaround can be removed when mpl is bumped to certain version
        except KeyError:
            colors = plt.rcParams['axes.color_cycle']
        else:
            colors = prop_cycle.by_key()['color']
        color_cycle = cycle(colors)
        step = 45
        with ImageComparison(self.testing_path, 'plot_ellipses_tangents.png',
                             style='classic', reltol=15) as ic:
            for angle in range(step, 360 + step - 1, step):
                x = dist * np.sin(np.radians(angle))
                y = dist * np.cos(np.radians(angle))
                ell = Ellipse(20, 10, 90, center=(x, y))
                if angle == 360:
                    outfile = ic.name
                else:
                    outfile = None
                fig = ell.plot_tangents((0, 0),
                                        fig=fig,
                                        color=next(color_cycle),
                                        print_angle=True,
                                        ellipse_name='E{:d}'.format(angle),
                                        outfile=outfile)
        # Test multi-station figure
        fig = None
        color_cycle = cycle(colors)
        with ImageComparison(self.testing_path,
                             'plot_ellipse_tangents_pts.png',
                             style='classic',
                             reltol=15) as ic:
            for angle in range(step, 360 + step - 1, step):
                x = dist * np.sin(np.radians(angle))
                y = dist * np.cos(np.radians(angle))
                ell = Ellipse(20, 10, 90)
                if angle == 360:
                    outfile = ic.name
                else:
                    outfile = None
                fig = ell.plot_tangents((x, y),
                                        fig=fig,
                                        color=next(color_cycle),
                                        print_angle=True,
                                        pt_name='pt{:d}'.format(angle),
                                        outfile=outfile)

    def test_read_uncert_ellipse(self):
        """
        Verify that confidence ellipse is properly read/calculated from nordic
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(
                os.path.join(self.testing_path, "sfile_highaccuracy"))
        event = cat[0]
        val = event.origins[0].origin_uncertainty
        hor_max = val['max_horizontal_uncertainty']
        hor_min = val['min_horizontal_uncertainty']
        azi_max = val['azimuth_max_horizontal_uncertainty']
        self.assertAlmostEqual(hor_max, 1120.674193646)
        self.assertAlmostEqual(hor_min, 451.762494786)
        self.assertAlmostEqual(azi_max, 167.9407699)


def _assert_similarity(event_1, event_2):
    """
    Raise AssertionError if testing similarity fails
    """
    similarity_output = _test_similarity(event_1, event_2)
    if similarity_output:
        raise AssertionError(similarity_output)


def _test_similarity(event_1, event_2):
    """
    Check the similarity of the components of obspy events, discounting
    resource IDs, which are not maintained in nordic files.

    Raise AssertionError if test fails

    :type event_1: obspy.core.event.Event
    :param event_1: First event
    :type event_2: obspy.core.event.Event
    :param event_2: Comparison event
    :type verbose: bool
    :param verbose: If true and fails will output why it fails.
    """
    # What None maps to.
    pick_default_mapper = {
        "polarity": "undecidable", "evaluation_mode": "manual"}
    # Check origins
    if len(event_1.origins) != len(event_2.origins):
        return False
    for ori_1, ori_2 in zip(event_1.origins, event_2.origins):
        for key in ori_1.keys():
            if key not in ["resource_id", "comments", "arrivals",
                           "method_id", "origin_uncertainty", "depth_type",
                           "quality", "creation_info", "evaluation_mode",
                           "depth_errors", "time_errors"]:
                if ori_1[key] != ori_2[key]:
                    return ('%s is not the same as %s for key %s' %
                            (ori_1[key], ori_2[key], key))
            elif key == "arrivals":
                if len(ori_1[key]) != len(ori_2[key]):
                    return ('%i is not the same as %i for key %s' %
                            (len(ori_1[key]), len(ori_2[key]), key))
                for arr_1, arr_2 in zip(ori_1[key], ori_2[key]):
                    for arr_key in arr_1.keys():
                        if arr_key not in ["resource_id", "pick_id",
                                           "distance"]:
                            if arr_1[arr_key] != arr_2[arr_key]:
                                return ('%s does not match %s for key %s' %
                                        (arr_1[arr_key], arr_2[arr_key],
                                         arr_key))
                    if arr_1["distance"] and round(
                            arr_1["distance"]) != round(arr_2["distance"]):
                        return ('%s does not match %s for key %s' %
                                (arr_1[arr_key], arr_2[arr_key],
                                 arr_key))
    # Check picks
    if len(event_1.picks) != len(event_2.picks):
        return 'Number of picks is not equal'
    for pick_1, pick_2 in zip(event_1.picks, event_2.picks):
        # Assuming same ordering of picks...
        for key in pick_1.keys():
            if key not in ["resource_id", "waveform_id"]:
                if pick_1[key] != pick_2[key]:
                    default = pick_default_mapper.get(key, None)
                    if pick_1[key] is None:
                        if pick_2[key] == default:
                            continue
                    elif pick_2[key] is None:
                        if pick_1[key] == default:
                            continue
                    return ('%s is not the same as %s for key %s' %
                            (pick_1[key], pick_2[key], key))
            elif key == "waveform_id":
                if pick_1[key].station_code != pick_2[key].station_code:
                    return 'Station codes do not match'
                if pick_1[key].channel_code[0] != pick_2[key].channel_code[0]:
                    return 'Channel codes do not match'
                if pick_1[key].channel_code[-1] !=\
                   pick_2[key].channel_code[-1]:
                    return 'Channel codes do not match'
    # Check amplitudes
    if not len(event_1.amplitudes) == len(event_2.amplitudes):
        return 'Not the same number of amplitudes'
    for amp_1, amp_2 in zip(event_1.amplitudes, event_2.amplitudes):
        # Assuming same ordering of amplitudes
        for key in amp_1.keys():
            if key not in ["resource_id", "pick_id", "waveform_id", "snr",
                           "magnitude_hint", 'type']:
                if not amp_1[key] == amp_2[key]:
                    return ("{0} is not the same as {1} for key "
                            "{2}\n{3}\n{4}".format(
                                amp_1[key], amp_2[key], key, amp_1, amp_2))
            elif key == "waveform_id":
                if pick_1[key].station_code != pick_2[key].station_code:
                    return 'Station codes do not match'
                if pick_1[key].channel_code[0] != pick_2[key].channel_code[0]:
                    return 'Channel codes do not match'
                if pick_1[key].channel_code[-1] !=\
                        pick_2[key].channel_code[-1]:
                    return 'Channel codes do not match'
            elif key in ["magnitude_hint", "type"]:
                # Reading back in will define both, but input event might have
                # None
                if amp_1[key] is not None:
                    if not amp_1[key] == amp_2[key]:
                        return ('%s is not the same as %s for key %s' %
                                (amp_1[key], amp_2[key], key))
    return None


def full_test_event():
    """
    Function to generate a basic, full test event
    """
    test_event = Event()
    test_event.origins.append(Origin(
        time=UTCDateTime("2012-03-26") + 1.2, latitude=45.0, longitude=25.0,
        depth=15000))
    test_event.event_descriptions.append(EventDescription())
    test_event.event_descriptions[0].text = 'LE'
    test_event.creation_info = CreationInfo(agency_id='TES')
    test_event.magnitudes.append(Magnitude(
        mag=0.1, magnitude_type='ML', creation_info=CreationInfo('TES'),
        origin_id=test_event.origins[0].resource_id))
    test_event.magnitudes.append(Magnitude(
        mag=0.5, magnitude_type='Mc', creation_info=CreationInfo('TES'),
        origin_id=test_event.origins[0].resource_id))
    test_event.magnitudes.append(Magnitude(
        mag=1.3, magnitude_type='Ms', creation_info=CreationInfo('TES'),
        origin_id=test_event.origins[0].resource_id))

    # Define the test pick
    _waveform_id_1 = WaveformStreamID(station_code='FOZ', channel_code='SHZ',
                                      network_code='NZ')
    _waveform_id_2 = WaveformStreamID(station_code='WTSZ', channel_code='BH1',
                                      network_code=' ')
    # Pick to associate with amplitude - 0
    test_event.picks = [
        Pick(waveform_id=_waveform_id_1, phase_hint='IAML',
             polarity='undecidable', time=UTCDateTime("2012-03-26") + 1.68,
             evaluation_mode="manual"),
        Pick(waveform_id=_waveform_id_1, onset='impulsive', phase_hint='PN',
             polarity='positive', time=UTCDateTime("2012-03-26") + 1.68,
             evaluation_mode="manual"),
        Pick(waveform_id=_waveform_id_1, phase_hint='IAML',
             polarity='undecidable', time=UTCDateTime("2012-03-26") + 1.68,
             evaluation_mode="manual"),
        Pick(waveform_id=_waveform_id_2, onset='impulsive', phase_hint='SG',
             polarity='undecidable', time=UTCDateTime("2012-03-26") + 1.72,
             evaluation_mode="manual"),
        Pick(waveform_id=_waveform_id_2, onset='impulsive', phase_hint='PN',
             polarity='undecidable', time=UTCDateTime("2012-03-26") + 1.62,
             evaluation_mode="automatic"),
        # Missing info shouldn't be an issue
        Pick(waveform_id=_waveform_id_2, onset=None, phase_hint='PN',
             polarity=None, time=UTCDateTime("2012-03-26") + 1.92,
             evaluation_mode=None),
        # Long-phase
        Pick(waveform_id=_waveform_id_2, onset='impulsive', phase_hint='PKiKP',
             polarity=None, time=UTCDateTime("2012-03-26") + 1.92,
             evaluation_mode=None),
    ]
    # Test a generic local magnitude amplitude pick
    test_event.amplitudes = [
        Amplitude(generic_amplitude=2.0, period=0.4,
                  pick_id=test_event.picks[0].resource_id,
                  waveform_id=test_event.picks[0].waveform_id, unit='m',
                  magnitude_hint='ML', category='point', type='AML'),
        Amplitude(generic_amplitude=10,
                  pick_id=test_event.picks[1].resource_id,
                  waveform_id=test_event.picks[1].waveform_id, type='END',
                  category='duration', unit='s', magnitude_hint='Mc',
                  snr=2.3),
        Amplitude(generic_amplitude=5.0, period=0.6,
                  pick_id=test_event.picks[2].resource_id,
                  waveform_id=test_event.picks[0].waveform_id, unit='m',
                  category='point', type='AML')]
    test_event.origins[0].arrivals = [
        Arrival(time_weight=0, phase=test_event.picks[1].phase_hint,
                pick_id=test_event.picks[1].resource_id),
        Arrival(time_weight=2, phase=test_event.picks[3].phase_hint,
                pick_id=test_event.picks[3].resource_id,
                backazimuth_residual=5, time_residual=0.2, distance=15,
                azimuth=25, takeoff_angle=10),
        Arrival(time_weight=2, phase=test_event.picks[4].phase_hint,
                pick_id=test_event.picks[4].resource_id,
                backazimuth_residual=5, time_residual=0.2, distance=15,
                azimuth=25, takeoff_angle=170),
        Arrival(time_weight=2, phase=test_event.picks[5].phase_hint,
                pick_id=test_event.picks[5].resource_id,
                backazimuth_residual=5, time_residual=0.2, distance=15,
                azimuth=25, takeoff_angle=170),
        Arrival(time_weight=0, phase=test_event.picks[6].phase_hint,
                pick_id=test_event.picks[6].resource_id,
                backazimuth_residual=5, time_residual=0.2, distance=15,
                azimuth=25, takeoff_angle=170),
    ]
    # Add in error info (line E)
    test_event.origins[0].quality = OriginQuality(
        standard_error=0.01, azimuthal_gap=36)
    # Origin uncertainty in Seisan is output as long-lat-depth, quakeML has
    # semi-major and semi-minor
    test_event.origins[0].origin_uncertainty = OriginUncertainty(
        confidence_ellipsoid=ConfidenceEllipsoid(
            semi_major_axis_length=3000, semi_minor_axis_length=1000,
            semi_intermediate_axis_length=2000, major_axis_plunge=20,
            major_axis_azimuth=100, major_axis_rotation=4))
    test_event.origins[0].time_errors = QuantityError(uncertainty=0.5)
    # Add in fault-plane solution info (line F) - Note have to check program
    # used to determine which fields are filled....
    test_event.focal_mechanisms.append(FocalMechanism(
        nodal_planes=NodalPlanes(nodal_plane_1=NodalPlane(
            strike=180, dip=20, rake=30, strike_errors=QuantityError(10),
            dip_errors=QuantityError(10), rake_errors=QuantityError(20))),
        method_id=ResourceIdentifier("smi:nc.anss.org/focalMechanism/FPFIT"),
        creation_info=CreationInfo(agency_id="NC"), misfit=0.5,
        station_distribution_ratio=0.8))
    # Need to test high-precision origin and that it is preferred origin.
    # Moment tensor includes another origin
    test_event.origins.append(Origin(
        time=UTCDateTime("2012-03-26") + 1.2, latitude=45.1, longitude=25.2,
        depth=14500))
    test_event.magnitudes.append(Magnitude(
        mag=0.1, magnitude_type='MW', creation_info=CreationInfo('TES'),
        origin_id=test_event.origins[-1].resource_id))
    # Moment tensors go with focal-mechanisms
    test_event.focal_mechanisms.append(FocalMechanism(
        moment_tensor=MomentTensor(
            derived_origin_id=test_event.origins[-1].resource_id,
            moment_magnitude_id=test_event.magnitudes[-1].resource_id,
            scalar_moment=100, tensor=Tensor(
                m_rr=100, m_tt=100, m_pp=10, m_rt=1, m_rp=20, m_tp=15),
            method_id=ResourceIdentifier(
                'smi:nc.anss.org/momentTensor/BLAH'))))
    return test_event


def suite():
    return unittest.makeSuite(TestNordicMethods, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests the SeismicPhase class.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import inspect
import os
import unittest

from obspy.core.util.misc import TemporaryWorkingDirectory
from obspy.taup import TauPyModel
from obspy.taup.tau_model import TauModel
from obspy.taup.seismic_phase import SeismicPhase
from obspy.taup.taup_create import build_taup_model


# Most generic way to get the data folder path.
DATA = os.path.join(os.path.dirname(os.path.abspath(
    inspect.getfile(inspect.currentframe()))), "data")


class TauPySeismicPhaseTestCase(unittest.TestCase):
    """
    Test suite for the SeismicPhase class.
    """
    def setUp(self):
        self.depth = 119
        self.tau_model = TauModel.from_file('iasp91').depth_correct(self.depth)

    def test_shoot_existing_ray_param(self):
        self.do_shoot_existing_ray_param_for_phase('P')
        self.do_shoot_existing_ray_param_for_phase('S')
        self.do_shoot_existing_ray_param_for_phase('p')
        self.do_shoot_existing_ray_param_for_phase('s')
        self.do_shoot_existing_ray_param_for_phase('PP')
        self.do_shoot_existing_ray_param_for_phase('SS')
        self.do_shoot_existing_ray_param_for_phase('PcP')
        self.do_shoot_existing_ray_param_for_phase('ScS')
        self.do_shoot_existing_ray_param_for_phase('PKP')
        self.do_shoot_existing_ray_param_for_phase('SKS')
        self.do_shoot_existing_ray_param_for_phase('PKIKP')
        self.do_shoot_existing_ray_param_for_phase('SKIKS')

    def do_shoot_existing_ray_param_for_phase(self, phase_name):
        phase = SeismicPhase(phase_name, self.tau_model)
        for i, ray_param in enumerate(phase.ray_param):
            max_rp_arrival = phase.shoot_ray(-1, ray_param)
            self.assertAlmostEqual(phase.dist[i], max_rp_arrival.purist_dist,
                                   delta=0.0001)
            self.assertAlmostEqual(phase.time[i], max_rp_arrival.time,
                                   delta=0.0001)

    def test_shoot_middle_ray_param(self):
        phase = SeismicPhase('P', self.tau_model)
        for i in range(phase.ray_param.shape[0] - 1):
            rp = (phase.ray_param[i] + phase.ray_param[i + 1]) / 2
            time_tol = abs(phase.time[i] - phase.time[i + 1])
            max_rp_arrival = phase.shoot_ray(-1, rp)
            self.assertAlmostEqual(phase.dist[i], max_rp_arrival.purist_dist,
                                   delta=0.1)
            self.assertAlmostEqual(phase.time[i], max_rp_arrival.time,
                                   delta=time_tol)
            self.assertAlmostEqual(phase.dist[i + 1],
                                   max_rp_arrival.purist_dist, delta=0.1)
            self.assertAlmostEqual(phase.time[i + 1], max_rp_arrival.time,
                                   delta=time_tol)

    def test_many_identically_named_phases(self):
        """
        Regression test to make sure obspy.taup works with models that
        produce many identically names seismic phases.
        """
        with TemporaryWorkingDirectory():
            folder = os.path.abspath(os.curdir)
            model_name = "smooth_geodynamic_model"
            build_taup_model(
                filename=os.path.join(DATA, model_name + ".tvel"),
                output_folder=folder, verbose=False)
            m = TauPyModel(os.path.join(folder, model_name + ".npz"))
        arr = m.get_ray_paths(172.8000, 46.762440693494824, ["SS"])
        self.assertGreater(len(arr), 10)


def suite():
    return unittest.makeSuite(TauPySeismicPhaseTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""
A waveform indexer collecting metadata from a file based waveform archive and
storing in into a standard SQL database.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import fnmatch
import os
import sys
import time

from obspy import read
from obspy.core.preview import create_preview
from obspy.core.util.base import _get_entry_points
from obspy.db.db import (WaveformChannel, WaveformFeatures, WaveformFile,
                         WaveformGaps, WaveformPath)


class WaveformFileCrawler(object):
    """
    A waveform file crawler.

    This class scans periodically all given paths for waveform files and
    collects them into a watch list.
    """
    def _update_or_insert(self, dataset):
        """
        Add a new file into or modifies existing file in database.
        """
        if len(dataset) < 1:
            return
        session = self.session()
        data = dataset[0]
        # check for duplicates
        if self.options.check_duplicates:
            query = session.query(WaveformFile, WaveformChannel, WaveformPath)
            query = query.filter(WaveformPath.id == WaveformFile.path_id)
            query = query.filter(WaveformFile.id == WaveformChannel.file_id)
            query = query.filter(WaveformPath.path != data['path'])
            query = query.filter(WaveformFile.file == data['file'])
            query = query.filter(WaveformChannel.network == data['network'])
            query = query.filter(WaveformChannel.station == data['station'])
            query = query.filter(WaveformChannel.location == data['location'])
            query = query.filter(WaveformChannel.channel == data['channel'])
            query = query.filter(WaveformChannel.starttime ==
                                 data['starttime'])
            query = query.filter(WaveformChannel.endtime == data['endtime'])
            if query.count() > 0:
                msg = "Duplicate entry '%s' in '%s'."
                self.log.error(msg % (data['file'], data['path']))
                return
        # fetch or create path
        try:
            # search for existing path
            query = session.query(WaveformPath)
            path = query.filter_by(path=data['path']).one()
        except Exception:
            # create new path entry
            path = WaveformPath(data)
            session.add(path)
        # search and delete existing file entry
        msg = "Inserted"
        if path.id is not None:
            # search for existing file
            query = session.query(WaveformFile)
            files = query.filter_by(path_id=path.id,
                                    file=data['file']).all()
            if files:
                msg = "Updated"
            # delete existing file entry and all related information
            for file in files:
                session.delete(file)
        # create new file entry
        file = WaveformFile(data)
        path.files.append(file)
        # add channel entries
        for data in dataset:
            # create new channel entry
            channel = WaveformChannel(data)
            file.channels.append(channel)
            # add gaps
            for gap in data['gaps']:
                channel.gaps.append(WaveformGaps(gap))
            # add features
            for feature in data['features']:
                channel.features.append(WaveformFeatures(feature))
        try:
            session.commit()
        except Exception as e:
            session.rollback()
            self.log.error(str(e))
        else:
            self.log.debug("%s '%s' in '%s'" % (msg, data['file'],
                                                data['path']))
        session.close()

    def _delete(self, path, file=None):
        """
        Remove a file or all files with a given path from the database.
        """
        session = self.session()
        if file:
            query = session.query(WaveformFile)
            query = query.filter(WaveformPath.path == path)
            query = query.filter(WaveformFile.file == file)
            query = query.filter(WaveformPath.archived is False)
            for file_obj in query:
                session.delete(file_obj)
            try:
                session.commit()
            except Exception as e:
                session.rollback()
                msg = "Error deleting file '%s' in '%s': %s"
                self.log.error(msg % (file, path, e))
            else:
                self.log.debug("Deleting file '%s' in '%s'" % (file, path))
        else:
            query = session.query(WaveformPath)
            query = query.filter(WaveformPath.path == path)
            query = query.filter(WaveformPath.archived is False)
            for path_obj in query:
                session.delete(path_obj)
            try:
                session.commit()
            except Exception as e:
                session.rollback()
                self.log.error("Error deleting path '%s': %s" % (path, e))
            else:
                self.log.debug("Deleting path '%s'" % (path))
        session.close()

    def _select(self, path=None):
        """
        Fetch entry from database.
        """
        session = self.session()
        if path:
            # check database for file entries in specific path
            result = session.query("file", "mtime").from_statement("""
                SELECT file, mtime
                FROM default_waveform_paths as p, default_waveform_files as f
                WHERE p.id=f.path_id
                AND p.path=:path""").params(path=path).all()
            result = dict(result)
        else:
            # get all path entries from database
            result = session.query("path").from_statement("""
                SELECT path FROM default_waveform_paths""").all()
            result = [r[0] for r in result]
        session.close()
        return result

    def get_features(self):
        return self.paths[self._root][1]

    features = property(get_features)

    def get_patterns(self):
        return self.paths[self._root][0]

    patterns = property(get_patterns)

    def has_pattern(self, file):
        """
        Checks if the file name fits to the preferred file pattern.
        """
        for pattern in self.patterns:
            if fnmatch.fnmatch(file, pattern):
                return True
        return False

    def _process_output_queue(self):
        try:
            dataset = self.output_queue.pop(0)
        except Exception:
            pass
        else:
            self._update_or_insert(dataset)

    def _process_log_queue(self):
        try:
            msg = self.log_queue.pop(0)
        except Exception:
            pass
        else:
            if msg.startswith('['):
                self.log.error(msg)
            else:
                self.log.debug(msg)

    def _reset_walker(self):
        """
        Resets the crawler parameters.
        """
        # break if options run_once is set and a run was completed already
        if self.options.run_once and \
                getattr(self, 'first_run_complete', False):
            # before shutting down make sure all queues are empty!
            while self.output_queue or self.work_queue:
                msg = 'Crawler stopped but waiting for empty queues to exit.'
                self.log.debug(msg)
                if self.log_queue:
                    msg = 'log_queue still has %s item(s)'
                    self.log.debug(msg % len(self.log_queue))
                    # Fetch items from the log queue
                    self._process_log_queue()
                    continue
                if self.output_queue:
                    msg = 'output_queue still has %s item(s)'
                    self.log.debug(msg % len(self.output_queue))
                    # try to finalize a single processed stream object from
                    # output queue
                    self._process_output_queue()
                    continue
                if self.work_queue:
                    msg = 'work_queue still has %s items'
                    self.log.debug(msg % len(self.work_queue))
                time.sleep(10)
            self.log.debug('Crawler stopped by option run_once.')
            sys.exit()
            return
        self.log.debug('Crawler restarted.')
        # reset attributes
        self._current_path = None
        self._current_files = []
        self._db_files = {}
        # get search paths for waveform crawler
        self._roots = list(self.paths.keys())
        self._root = self._roots.pop(0)
        # create new walker
        self._walker = os.walk(self._root, topdown=True, followlinks=True)
        # clean up paths
        if self.options.cleanup:
            paths = self._select()
            for path in paths:
                if not os.path.isdir(path):
                    # no path in filesystem
                    self._delete(path)
                elif not self._select(path):
                    # empty path in database
                    self._delete(path)
        # logging
        self.log.debug("Crawling root '%s' ..." % self._root)
        self.first_run_complete = True

    def _step_walker(self):
        """
        Steps current walker object to the next directory.
        """
        # try to fetch next directory
        try:
            root, dirs, files = next(self._walker)
        except StopIteration:
            # finished cycling through all directories in current walker
            # try get next crawler search path
            try:
                self._root = self._roots.pop()
            except IndexError:
                # a whole cycle has been done
                # reset everything
                self._reset_walker()
                return
            # reset attributes
            self._current_path = None
            self._current_files = []
            self._db_files = {}
            # create new walker
            self._walker = os.walk(self._root, topdown=True, followlinks=True)
            # logging
            self.log.debug("Crawling root '%s' ..." % self._root)
            return
        # remove files or paths starting with a dot
        if self.options.skip_dots:
            for file in files:
                if file.startswith('.'):
                    files.remove(file)
            for dir in dirs:
                if dir.startswith('.'):
                    dirs.remove(dir)
        self._current_path = root
        self._current_files = files
        # logging
        self.log.debug("Scanning path '%s' ..." % self._current_path)
        # get all database entries for current path
        self._db_files = self._select(self._current_path)

    def _prepare_paths(self, paths):
        out = {}
        for path in paths:
            # strip features
            if '#' in path:
                parts = path.split('#')
                path = parts[0]
                features = parts[1:]
            else:
                features = []
            # strip patterns
            if '=' in path:
                path, patterns = path.split('=', 1)
                if ' ' in patterns:
                    patterns = patterns.split(' ')
                else:
                    patterns = [patterns.strip()]
            else:
                patterns = ['*.*']
            # normalize and absolute path name
            path = os.path.normpath(os.path.abspath(path))
            # check path
            if not os.path.isdir(path):
                self.log.warn("Skipping inaccessible path '%s' ..." % path)
                continue
            out[path] = (patterns, features)
        return out

    def iterate(self):
        """
        Handles exactly one directory.
        """
        # skip if service is not running
        # be aware that the processor pool is still active waiting for work
        if not self.running:
            return
        # skip if input queue is full
        if len(self.input_queue) > self.options.number_of_cpus:
            return
        # try to finalize a single processed stream object from output queue
        self._process_output_queue()
        # Fetch items from the log queue
        self._process_log_queue()
        # walk through directories and files
        try:
            file = self._current_files.pop(0)
        except IndexError:
            # file list is empty
            # clean up not existing files in current path
            if self.options.cleanup:
                for file in self._db_files.keys():
                    self._delete(self._current_path, file)
            # jump into next directory
            self._step_walker()
            return
        # skip file with wrong pattern
        if not self.has_pattern(file):
            return
        # process a single file
        path = self._current_path
        filepath = os.path.join(path, file)
        # get file stats
        try:
            stats = os.stat(filepath)
            mtime = int(stats.st_mtime)
        except Exception as e:
            self.log.error(str(e))
            return
        # check if recent
        if self.options.recent:
            # skip older files
            if time.time() - mtime > 60 * 60 * self.options.recent:
                try:
                    db_file_mtime = self._db_files.pop(file)
                except Exception:
                    pass
                return
        # option force-reindex set -> process file regardless if already in
        # database or recent or whatever
        if self.options.force_reindex:
            self.input_queue[filepath] = (path, file, self.features)
            return
        # compare with database entries
        if file not in self._db_files.keys():
            # file does not exists in database -> add file
            self.input_queue[filepath] = (path, file, self.features)
            return
        # file is already in database
        # -> remove from file list so it won't be deleted on database cleanup
        try:
            db_file_mtime = self._db_files.pop(file)
        except Exception:
            return
        # -> compare modification times of current file with database entry
        if mtime == db_file_mtime:
            return
        # modification time differs -> update file
        self.input_queue[filepath] = (path, file, self.features)


def worker(_i, input_queue, work_queue, output_queue, log_queue, mappings={}):
    try:
        # fetch and initialize all possible waveform feature plug-ins
        all_features = {}
        for (key, ep) in _get_entry_points('obspy.db.feature').items():
            try:
                # load plug-in
                cls = ep.load()
                # initialize class
                func = cls().process
            except Exception as e:
                msg = 'Could not initialize feature %s. (%s)'
                log_queue.append(msg % (key, str(e)))
                continue
            all_features[key] = {}
            all_features[key]['run'] = func
            try:
                all_features[key]['indexer_kwargs'] = cls['indexer_kwargs']
            except Exception:
                all_features[key]['indexer_kwargs'] = {}
        # loop through input queue
        while True:
            # fetch a unprocessed item
            try:
                filepath, (path, file, features) = input_queue.popitem()
            except Exception:
                continue
            # skip item if already in work queue
            if filepath in work_queue:
                continue
            work_queue.append(filepath)
            # get additional kwargs for read method from waveform plug-ins
            kwargs = {'verify_chksum': False}
            for feature in features:
                if feature not in all_features:
                    log_queue.append('%s: Unknown feature %s' % (filepath,
                                                                 feature))
                    continue
                kwargs.update(all_features[feature]['indexer_kwargs'])
            # read file and get file stats
            try:
                stats = os.stat(filepath)
                stream = read(filepath, **kwargs)
                # get gap and overlap information
                gap_list = stream.get_gaps()
                # merge channels and replace gaps/overlaps with 0 to prevent
                # generation of masked arrays
                stream.merge(fill_value=0)
            except Exception as e:
                msg = '[Reading stream] %s: %s'
                log_queue.append(msg % (filepath, e))
                try:
                    work_queue.remove(filepath)
                except Exception:
                    pass
                continue
            # build up dictionary of gaps and overlaps for easier lookup
            gap_dict = {}
            for gap in gap_list:
                id = '.'.join(gap[0:4])
                temp = {
                    'gap': gap[6] >= 0,
                    'starttime': gap[4].datetime,
                    'endtime': gap[5].datetime,
                    'samples': abs(gap[7])
                }
                gap_dict.setdefault(id, []).append(temp)
            # loop through traces
            dataset = []
            for trace in stream:
                result = {}
                # general file information
                result['mtime'] = int(stats.st_mtime)
                result['size'] = stats.st_size
                result['path'] = path
                result['file'] = file
                result['filepath'] = filepath
                # trace information
                result['format'] = trace.stats._format
                result['station'] = trace.stats.station
                result['location'] = trace.stats.location
                result['channel'] = trace.stats.channel
                result['network'] = trace.stats.network
                result['starttime'] = trace.stats.starttime.datetime
                result['endtime'] = trace.stats.endtime.datetime
                result['calib'] = trace.stats.calib
                result['npts'] = trace.stats.npts
                result['sampling_rate'] = trace.stats.sampling_rate
                # check for any id mappings
                if trace.id in mappings:
                    old_id = trace.id
                    for mapping in mappings[old_id]:
                        if trace.stats.starttime and \
                           trace.stats.starttime > mapping['endtime']:
                            continue
                        if trace.stats.endtime and \
                           trace.stats.endtime < mapping['starttime']:
                            continue
                        result['network'] = mapping['network']
                        result['station'] = mapping['station']
                        result['location'] = mapping['location']
                        result['channel'] = mapping['channel']
                        msg = "Mapping '%s' to '%s.%s.%s.%s'" % \
                            (old_id, mapping['network'], mapping['station'],
                             mapping['location'], mapping['channel'])
                        log_queue.append(msg)
                # gaps/overlaps for current trace
                result['gaps'] = gap_dict.get(trace.id, [])
                # apply feature functions
                result['features'] = []
                for key in features:
                    if key not in all_features:
                        continue
                    try:
                        # run plug-in and update results
                        temp = all_features[key]['run'](trace)
                        for key, value in temp.items():
                            result['features'].append({'key': key,
                                                       'value': value})
                    except Exception as e:
                        msg = '[Processing feature] %s: %s'
                        log_queue.append(msg % (filepath, e))
                        continue
                # generate preview of trace
                result['preview'] = None
                if '.LOG.L.' not in file or trace.stats.channel != 'LOG':
                    # create previews only for non-log files (see issue #400)
                    try:
                        trace = create_preview(trace, 30)
                        result['preview'] = trace.data.dumps()
                    except ValueError:
                        pass
                    except Exception as e:
                        msg = '[Creating preview] %s: %s'
                        log_queue.append(msg % (filepath, e))
                # update dataset
                dataset.append(result)
            del stream
            # return results to main loop
            try:
                output_queue.append(dataset)
            except Exception:
                pass
            try:
                work_queue.remove(filepath)
            except Exception:
                pass
    except KeyboardInterrupt:
        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The gse2.bulletin test suite.

:author:
    EOST (École et Observatoire des Sciences de la Terre)
:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import os
import unittest
import warnings

from obspy.core.event import read_events
from obspy.core.utcdatetime import UTCDateTime
from obspy.io.gse2.bulletin import _read_gse2, GSE2BulletinSyntaxError
from obspy.core.event.header import OriginDepthType, PickPolarity
from obspy.core.inventory.inventory import read_inventory


class BulletinTestCase(unittest.TestCase):
    """
    Test suite for obspy.io.gse2.bulletin
    """

    def setUp(self):
        # directory where the test files are located
        self.path = os.path.join(os.path.dirname(__file__), 'data/bulletin')

    def test_catalog(self):
        """
        Test Catalog object.
        """
        filename = os.path.join(self.path, 'gse_2.0_standard.txt')
        catalog = _read_gse2(filename)
        self.assertEqual(catalog.resource_id, 'smi:local/event/evid')
        self.assertEqual(len(catalog.comments), 1)
        comment = catalog.comments[0]
        text = ("Reviewed Event Bulletin (REB) of the GSE_IDC for January 16, "
                "1995")
        self.assertEqual(comment.text, text)
        self.assertEqual(comment.resource_id.id[:10], 'smi:local/')

    def test_event(self):
        """
        Test Event object.
        """
        filename = os.path.join(self.path, 'gse_2.0_standard.txt')
        catalog = _read_gse2(filename)
        self.assertEqual(len(catalog), 1)
        event = catalog[0]
        self.assertEqual(event.resource_id, 'smi:local/event/280435')
        self.assertEqual(event.event_type, 'earthquake')
        self.assertEqual(event.event_type_certainty, "known")
        self.assertNotEqual(event.creation_info, None)
        self.assertEqual(event.preferred_origin_id, 'smi:local/origin/282672')
        self.assertEqual(event.preferred_magnitude_id,
                         'smi:local/origin/282672/magnitude/0')
        # event descriptions
        self.assertEqual(len(event.event_descriptions), 1)
        event_description = event.event_descriptions[0]
        self.assertEqual(
            event_description.text,
            'GREECE-ALBANIA BORDER REGION')
        self.assertEqual(event_description.type, 'region name')
        # comments
        self.assertEqual(len(event.comments), 1)
        comment = event.comments[0]
        self.assertEqual(comment.text, 'GSE2.0:evtype=ke')
        self.assertEqual(comment.resource_id.id[:10], 'smi:local/')

    def test_origin(self):
        """
        Test Origin object.
        """
        filename = os.path.join(self.path, 'gse_2.0_standard.txt')
        catalog = _read_gse2(filename)
        self.assertEqual(len(catalog), 1)
        self.assertEqual(len(catalog[0].origins), 1)
        origin = catalog[0].origins[0]
        self.assertEqual(origin.resource_id, 'smi:local/origin/282672')
        self.assertEqual(origin.time, UTCDateTime('1995-01-16T07:26:52.4'))
        self.assertEqual(origin.time_errors.uncertainty, 12.69)
        self.assertEqual(origin.latitude, 39.45)
        self.assertEqual(origin.latitude_errors.uncertainty, None)
        self.assertEqual(origin.longitude, 20.44)
        self.assertEqual(origin.longitude_errors.uncertainty, None)
        self.assertEqual(origin.depth, 66800)
        self.assertEqual(origin.depth_errors.uncertainty, 83800)
        self.assertEqual(origin.depth_type, OriginDepthType('from location'))
        self.assertEqual(origin.time_fixed, False)
        self.assertEqual(origin.epicenter_fixed, False)
        self.assertEqual(origin.reference_system_id, None)
        self.assertEqual(origin.method_id, 'smi:local/method/inversion')
        self.assertEqual(origin.earth_model_id, None)
        self.assertEqual(origin.origin_type, None)
        self.assertEqual(origin.region, None)
        self.assertEqual(origin.evaluation_mode, "manual")
        self.assertEqual(origin.evaluation_status, None)
        # quality
        self.assertEqual(origin.quality.associated_phase_count, 9)
        self.assertEqual(origin.quality.used_phase_count, 9)
        self.assertEqual(origin.quality.associated_station_count, 8)
        self.assertEqual(origin.quality.used_station_count, 8)
        self.assertEqual(origin.quality.depth_phase_count, None)
        self.assertEqual(origin.quality.standard_error, 0.53)
        self.assertEqual(origin.quality.azimuthal_gap, 322)
        self.assertEqual(origin.quality.secondary_azimuthal_gap, None)
        self.assertEqual(origin.quality.ground_truth_level, None)
        self.assertEqual(origin.quality.minimum_distance, 10.56)
        self.assertEqual(origin.quality.maximum_distance, 78.21)
        self.assertEqual(origin.quality.median_distance, None)
        # origin uncertainty
        u = origin.origin_uncertainty
        self.assertEqual(u.horizontal_uncertainty, None)
        self.assertEqual(u.min_horizontal_uncertainty, 83700)
        self.assertEqual(u.max_horizontal_uncertainty, 93600)
        self.assertEqual(u.azimuth_max_horizontal_uncertainty, 27)
        self.assertEqual(u.confidence_ellipsoid, None)
        self.assertEqual(u.preferred_description, 'uncertainty ellipse')
        self.assertEqual(u.confidence_level, None)
        # creation info
        self.assertEqual(origin.creation_info.author, 'GSE_IDC')
        # comments
        self.assertEqual(len(origin.comments), 0)
        # composite times
        self.assertEqual(len(origin.composite_times), 0)

    def test_pick(self):
        """
        Test Pick object.
        """
        filename = os.path.join(self.path, 'gse_2.0_standard.txt')
        catalog = _read_gse2(filename)
        self.assertEqual(len(catalog), 1)
        picks = catalog[0].picks
        self.assertEqual(len(picks), 9)
        # Test first Pick
        pick_1 = picks[0]
        self.assertEqual(pick_1.resource_id, 'smi:local/pick/3586432')
        self.assertEqual(pick_1.time, UTCDateTime('1995-01-16T07:29:20.7'))
        # WaveformStreamId
        waveform_1 = pick_1.waveform_id
        self.assertEqual(waveform_1.network_code, 'XX')
        self.assertEqual(waveform_1.station_code, 'GERES')
        self.assertEqual(waveform_1.channel_code, None)
        self.assertEqual(waveform_1.location_code, None)
        self.assertEqual(waveform_1.resource_uri, None)
        self.assertEqual(pick_1.filter_id, None)
        self.assertEqual(pick_1.method_id, None)
        self.assertEqual(pick_1.horizontal_slowness, 13.8)
        self.assertEqual(pick_1.backazimuth, 163.7)
        self.assertEqual(pick_1.slowness_method_id, None)
        self.assertEqual(pick_1.onset, 'emergent')
        self.assertEqual(pick_1.phase_hint, 'P')
        self.assertEqual(pick_1.polarity, None)
        self.assertEqual(pick_1.evaluation_mode, 'manual')
        self.assertEqual(pick_1.evaluation_status, None)
        self.assertNotEqual(pick_1.creation_info, None)
        self.assertEqual(len(pick_1.comments), 0)
        # Test second Pick
        pick_2 = picks[1]
        self.assertEqual(pick_2.resource_id, 'smi:local/pick/3586513')
        self.assertEqual(pick_2.time, UTCDateTime('1995-01-16T07:31:17.5'))
        # WaveformStreamId
        waveform_2 = pick_2.waveform_id
        self.assertEqual(waveform_2.network_code, 'XX')
        self.assertEqual(waveform_2.station_code, 'GERES')
        self.assertEqual(waveform_2.channel_code, None)
        self.assertEqual(waveform_2.location_code, None)
        self.assertEqual(waveform_2.resource_uri, None)
        self.assertEqual(pick_2.filter_id, None)
        self.assertEqual(pick_2.method_id, None)
        self.assertEqual(pick_2.horizontal_slowness, 23.4)
        self.assertEqual(pick_2.backazimuth, 153.4)
        self.assertEqual(pick_2.slowness_method_id, None)
        self.assertEqual(pick_2.onset, None)
        self.assertEqual(pick_2.phase_hint, 'S')
        self.assertEqual(pick_2.polarity, PickPolarity.POSITIVE)
        self.assertEqual(pick_2.evaluation_mode, None)
        self.assertEqual(pick_2.evaluation_status, None)
        self.assertNotEqual(pick_2.creation_info, None)
        self.assertEqual(len(pick_2.comments), 0)

    def test_arrival(self):
        """
        Test Arrival object.
        """
        filename = os.path.join(self.path, 'gse_2.0_standard.txt')
        catalog = _read_gse2(filename)
        self.assertEqual(len(catalog), 1)
        self.assertEqual(len(catalog[0].origins), 1)
        arrivals = catalog[0].origins[0].arrivals
        self.assertEqual(len(arrivals), 9)
        # Test first Arrival
        arrival_1 = arrivals[0]
        self.assertEqual(
            arrival_1.resource_id, 'smi:local/origin/282672/arrival/3586432')
        self.assertEqual(arrival_1.pick_id, 'smi:local/pick/3586432')
        self.assertEqual(arrival_1.phase, 'P')
        self.assertEqual(arrival_1.time_correction, None)
        self.assertEqual(arrival_1.azimuth, 150.3)
        self.assertEqual(arrival_1.distance, 10.56)
        self.assertEqual(arrival_1.takeoff_angle, None)
        self.assertEqual(arrival_1.time_residual, -0.2)
        self.assertEqual(arrival_1.horizontal_slowness_residual, 0.1)
        self.assertEqual(arrival_1.backazimuth_residual, 13.4)
        self.assertEqual(arrival_1.time_weight, 1)
        self.assertEqual(arrival_1.backazimuth_weight, None)
        self.assertEqual(arrival_1.horizontal_slowness_weight, None)
        self.assertEqual(arrival_1.earth_model_id, None)
        self.assertNotEqual(arrival_1.creation_info, None)
        self.assertEqual(len(arrival_1.comments), 0)
        # Test second Arrival
        arrival_2 = arrivals[1]
        self.assertEqual(
            arrival_2.resource_id, 'smi:local/origin/282672/arrival/3586513')
        self.assertEqual(arrival_2.pick_id, 'smi:local/pick/3586513')
        self.assertEqual(arrival_2.phase, 'S')
        self.assertEqual(arrival_2.time_correction, None)
        self.assertEqual(arrival_2.azimuth, 150.3)
        self.assertEqual(arrival_2.distance, 10.56)
        self.assertEqual(arrival_2.takeoff_angle, None)
        self.assertEqual(arrival_2.time_residual, -0.6)
        self.assertEqual(arrival_2.horizontal_slowness_residual, -1.0)
        self.assertEqual(arrival_2.backazimuth_residual, 3.1)
        self.assertEqual(arrival_2.time_weight, None)
        self.assertEqual(arrival_2.backazimuth_weight, 1)
        self.assertEqual(arrival_2.horizontal_slowness_weight, None)
        self.assertEqual(arrival_2.earth_model_id, None)
        self.assertNotEqual(arrival_2.creation_info, None)
        self.assertEqual(len(arrival_2.comments), 0)

    def test_magnitude(self):
        """
        Test Magnitude object.
        """
        filename = os.path.join(self.path, 'gse_2.0_standard.txt')
        catalog = _read_gse2(filename)
        self.assertEqual(len(catalog), 1)
        magnitudes = catalog[0].magnitudes
        self.assertEqual(len(magnitudes), 2)
        # Test first Magnitude
        mag_1 = magnitudes[0]
        self.assertEqual(
            mag_1.resource_id, 'smi:local/origin/282672/magnitude/0')
        self.assertEqual(mag_1.mag, 3.6)
        self.assertEqual(mag_1.mag_errors.uncertainty, 0.2)
        self.assertEqual(mag_1.magnitude_type, 'mb')
        self.assertEqual(mag_1.origin_id, 'smi:local/origin/282672')
        self.assertEqual(mag_1.method_id, None)
        self.assertEqual(mag_1.station_count, 3)
        self.assertEqual(mag_1.azimuthal_gap, None)
        self.assertEqual(mag_1.evaluation_mode, None)
        self.assertEqual(mag_1.evaluation_status, None)
        self.assertNotEqual(mag_1.creation_info, None)
        self.assertEqual(len(mag_1.comments), 0)
        self.assertEqual(len(mag_1.station_magnitude_contributions), 0)
        # Test second Magnitude
        mag_2 = magnitudes[1]
        self.assertEqual(
            mag_2.resource_id, 'smi:local/origin/282672/magnitude/1')
        self.assertEqual(mag_2.mag, 4.0)
        self.assertEqual(mag_2.mag_errors.uncertainty, None)
        self.assertEqual(mag_2.magnitude_type, 'ML')
        self.assertEqual(mag_2.origin_id, 'smi:local/origin/282672')
        self.assertEqual(mag_2.method_id, None)
        self.assertEqual(mag_2.station_count, 1)
        self.assertEqual(mag_2.azimuthal_gap, None)
        self.assertEqual(mag_2.evaluation_mode, None)
        self.assertEqual(mag_2.evaluation_status, None)
        self.assertNotEqual(mag_2.creation_info, None)
        self.assertEqual(len(mag_1.comments), 0)
        self.assertEqual(len(mag_1.station_magnitude_contributions), 0)

    def test_station_magnitude(self):
        """
        Test StationMagnitude object.
        """
        filename = os.path.join(self.path, 'gse_2.0_standard.txt')
        catalog = _read_gse2(filename)
        self.assertEqual(len(catalog), 1)
        station_magnitudes = catalog[0].station_magnitudes
        self.assertEqual(len(station_magnitudes), 4)
        # Test first StationMagnitude
        sta_mag_1 = station_magnitudes[0]
        self.assertEqual(
            sta_mag_1.resource_id, 'smi:local/magnitude/station/3586432/0')
        self.assertEqual(sta_mag_1.origin_id, 'smi:local/origin/282672')
        self.assertEqual(sta_mag_1.mag, 4.0)
        self.assertEqual(sta_mag_1.station_magnitude_type, 'ML')
        self.assertEqual(sta_mag_1.amplitude_id, 'smi:local/amplitude/3586432')
        self.assertEqual(sta_mag_1.method_id, None)
        self.assertEqual(sta_mag_1.waveform_id, None)
        self.assertNotEqual(sta_mag_1.creation_info, None)
        self.assertEqual(len(sta_mag_1.comments), 0)
        # Test second StationMagnitude
        sta_mag_2 = station_magnitudes[1]
        self.assertEqual(
            sta_mag_2.resource_id, 'smi:local/magnitude/station/3586555/0')
        self.assertEqual(sta_mag_2.origin_id, 'smi:local/origin/282672')
        self.assertEqual(sta_mag_2.mag, 3.7)
        self.assertEqual(sta_mag_2.station_magnitude_type, 'mb')
        self.assertEqual(sta_mag_2.amplitude_id, 'smi:local/amplitude/3586555')
        self.assertEqual(sta_mag_2.method_id, None)
        self.assertEqual(sta_mag_2.waveform_id, None)
        self.assertNotEqual(sta_mag_2.creation_info, None)
        self.assertEqual(len(sta_mag_2.comments), 0)

    def test_amplitude(self):
        """
        Test Amplitude object.
        """
        filename = os.path.join(self.path, 'gse_2.0_standard.txt')
        catalog = _read_gse2(filename)
        self.assertEqual(len(catalog), 1)
        amplitudes = catalog[0].amplitudes
        self.assertEqual(len(amplitudes), 6)
        # Test first amplitude
        amplitude_1 = amplitudes[0]
        self.assertEqual(
            amplitude_1.resource_id, 'smi:local/amplitude/3586432')
        self.assertEqual(amplitude_1.generic_amplitude, 0.6)
        self.assertEqual(amplitude_1.type, None)
        self.assertEqual(amplitude_1.category, None)
        self.assertEqual(amplitude_1.unit, None)
        self.assertEqual(amplitude_1.method_id, None)
        self.assertEqual(amplitude_1.period, 0.3)
        self.assertEqual(amplitude_1.snr, 6.8)
        self.assertEqual(amplitude_1.time_window, None)
        self.assertEqual(amplitude_1.pick_id, 'smi:local/pick/3586432')
        # WaveformStreamId
        waveform_1 = amplitude_1.waveform_id
        self.assertEqual(waveform_1.network_code, 'XX')
        self.assertEqual(waveform_1.station_code, 'GERES')
        self.assertEqual(waveform_1.channel_code, None)
        self.assertEqual(waveform_1.location_code, None)
        self.assertEqual(waveform_1.resource_uri, None)
        self.assertEqual(amplitude_1.filter_id, None)
        self.assertEqual(amplitude_1.scaling_time, None)
        self.assertEqual(amplitude_1.magnitude_hint, 'ML')
        self.assertEqual(amplitude_1.evaluation_mode, None)
        self.assertEqual(amplitude_1.evaluation_status, None)
        self.assertNotEqual(amplitude_1.creation_info, None)
        self.assertEqual(len(amplitude_1.comments), 0)
        # Test second amplitude
        amplitude_2 = amplitudes[1]
        self.assertEqual(
            amplitude_2.resource_id, 'smi:local/amplitude/3586513')
        self.assertEqual(amplitude_2.generic_amplitude, 2.9)
        self.assertEqual(amplitude_2.type, None)
        self.assertEqual(amplitude_2.category, None)
        self.assertEqual(amplitude_2.unit, None)
        self.assertEqual(amplitude_2.method_id, None)
        self.assertEqual(amplitude_2.period, 0.6)
        self.assertEqual(amplitude_2.snr, 4.9)
        self.assertEqual(amplitude_2.time_window, None)
        self.assertEqual(amplitude_2.pick_id, 'smi:local/pick/3586513')
        # WaveformStreamId
        waveform_2 = amplitude_2.waveform_id
        self.assertEqual(waveform_2.network_code, 'XX')
        self.assertEqual(waveform_2.station_code, 'GERES')
        self.assertEqual(waveform_2.channel_code, None)
        self.assertEqual(waveform_2.location_code, None)
        self.assertEqual(waveform_2.resource_uri, None)
        self.assertEqual(amplitude_2.filter_id, None)
        self.assertEqual(amplitude_2.scaling_time, None)
        self.assertEqual(amplitude_2.magnitude_hint, None)
        self.assertEqual(amplitude_2.evaluation_mode, None)
        self.assertEqual(amplitude_2.evaluation_status, None)
        self.assertNotEqual(amplitude_2.creation_info, None)
        self.assertEqual(len(amplitude_2.comments), 0)

    def test_several_events(self):
        """
        Test with several events.
        """
        filename = os.path.join(self.path, 'gse_2.0_2_events.txt')
        catalog = _read_gse2(filename)
        self.assertEqual(len(catalog), 2)
        # Test firt event
        event_1 = catalog[0]
        self.assertEqual(event_1.resource_id, 'smi:local/event/280435')
        self.assertEqual(len(event_1.event_descriptions), 1)
        self.assertEqual(
            event_1.event_descriptions[0].text,
            'GREECE-ALBANIA BORDER REGION')
        self.assertEqual(len(event_1.comments), 1)
        self.assertEqual(len(event_1.picks), 9)
        self.assertEqual(len(event_1.amplitudes), 6)
        self.assertEqual(len(event_1.origins), 1)
        self.assertEqual(len(event_1.magnitudes), 2)
        self.assertEqual(len(event_1.station_magnitudes), 4)
        # Test second event
        event_2 = catalog[1]
        self.assertEqual(event_2.resource_id, 'smi:local/event/280436')
        self.assertEqual(len(event_2.event_descriptions), 1)
        self.assertEqual(
            event_2.event_descriptions[0].text,
            'VANCOUVER ISLAND REGION')
        self.assertEqual(len(event_2.comments), 1)
        self.assertEqual(len(event_2.picks), 7)
        self.assertEqual(len(event_2.amplitudes), 5)
        self.assertEqual(len(event_2.origins), 1)
        self.assertEqual(len(event_2.magnitudes), 1)
        self.assertEqual(len(event_2.station_magnitudes), 2)

    def test_parameters(self):
        filename = os.path.join(self.path, 'gse_2.0_standard.txt')
        fields = {
            'line_1': {
                'time': slice(0, 21),
                'time_fixf': slice(22, 23),
                'lat': slice(25, 33),
                'lon': slice(34, 43),
                'epicenter_fixf': slice(44, 45),
                'depth': slice(47, 52),
                'depth_fixf': slice(53, 54),
                'n_def': slice(56, 60),
                'n_sta': slice(61, 65),
                'gap': slice(66, 69),
                'mag_type_1': slice(71, 73),
                'mag_1': slice(73, 77),
                'mag_n_sta_1': slice(78, 80),
                'mag_type_2': slice(82, 84),
                'mag_2': slice(84, 88),
                'mag_n_sta_2': slice(89, 91),
                'mag_type_3': slice(93, 95),
                'mag_3': slice(95, 99),
                'mag_n_sta_3': slice(100, 102),
                'author': slice(104, 112),
                'id': slice(114, 122),
            },
            'line_2': {
                'rms': slice(5, 10),
                'ot_error': slice(15, 21),
                's_major': slice(25, 31),
                's_minor': slice(32, 38),
                'az': slice(40, 43),
                'depth_err': slice(49, 54),
                'min_dist': slice(56, 62),
                'max_dist': slice(63, 69),
                'mag_err_1': slice(74, 77),
                'mag_err_2': slice(85, 88),
                'mag_err_3': slice(96, 99),
                'antype': slice(104, 105),
                'loctype': slice(106, 107),
                'evtype': slice(108, 110),
            },
            'arrival': {
                'sta': slice(0, 5),
                'dist': slice(6, 12),
                'ev_az': slice(13, 18),
                'picktype': slice(19, 20),
                'direction': slice(20, 21),
                'detchar': slice(21, 22),
                'phase': slice(23, 30),
                'time': slice(31, 52),
                't_res': slice(53, 58),
                'azim': slice(59, 64),
                'az_res': slice(65, 71),
                'slow': slice(72, 77),
                's_res': slice(78, 83),
                't_def': slice(84, 85),
                'a_def': slice(85, 86),
                's_def': slice(86, 87),
                'snr': slice(88, 93),
                'amp': slice(94, 103),
                'per': slice(104, 109),
                'mag_type_1': slice(110, 112),
                'mag_1': slice(112, 116),
                'mag_type_2': slice(117, 119),
                'mag_2': slice(119, 123),
                'id': slice(124, 132),
            },
        }
        catalog = _read_gse2(filename, None, 'TZ', '00', 'SHZ', 'quakeml:idc',
                             fields, False, 'AGE')
        self.assertEqual(len(catalog), 1)
        # Test Comment ResourceIdentifier
        self.assertEqual(len(catalog.comments), 1)
        comment = catalog.comments[0]
        self.assertEqual(comment.resource_id.id[:12], 'quakeml:idc/')
        # Test Event ResourceIdentifier
        event = catalog[0]
        self.assertEqual(event.resource_id, 'quakeml:idc/event/280435')
        self.assertEqual(event.creation_info.agency_id, 'AGE')
        # Test Origin ResourceIdentifier
        self.assertEqual(len(event.origins), 1)
        origin = event.origins[0]
        self.assertEqual(origin.resource_id, 'quakeml:idc/origin/282672')
        self.assertEqual(origin.method_id, 'quakeml:idc/method/inversion')
        # Test Pick ResourceIdentifier
        self.assertEqual(len(event.picks), 9)
        pick = event.picks[0]
        self.assertEqual(pick.resource_id, 'quakeml:idc/pick/3586432')
        # Test Arrival ResourceIdentifier
        self.assertEqual(len(origin.arrivals), 9)
        arrival = origin.arrivals[0]
        self.assertEqual(
            arrival.resource_id, 'quakeml:idc/origin/282672/arrival/3586432')
        # Test Magnitude ResourceIdentifier
        self.assertEqual(len(event.magnitudes), 2)
        magnitude = event.magnitudes[0]
        self.assertEqual(
            magnitude.resource_id, 'quakeml:idc/origin/282672/magnitude/0')
        # Test StationMagnitude ResourceIdentifier
        self.assertEqual(len(event.station_magnitudes), 4)
        sta_mag = event.station_magnitudes[0]
        self.assertEqual(
            sta_mag.resource_id, 'quakeml:idc/magnitude/station/3586432/0')
        # Test Amplitude ResourceIdentifier
        self.assertEqual(len(event.amplitudes), 6)
        amplitude = event.amplitudes[0]
        self.assertEqual(
            amplitude.resource_id, 'quakeml:idc/amplitude/3586432')
        # Test network code
        waveform = pick.waveform_id
        self.assertEqual(waveform.network_code, 'TZ')
        self.assertEqual(waveform.channel_code, 'SHZ')
        self.assertEqual(waveform.location_code, '00')

    def test_non_standard_format(self):
        """
        Test non-standard GSE2 format which can normally be parsed too.
        """
        filename = os.path.join(self.path, 'gse_2.0_non_standard.txt')
        fields = {
            'line_1': {
                'author': slice(105, 113),
                'id': slice(114, 123),
            },
            'line_2': {
                'az': slice(40, 46),
                'antype': slice(105, 106),
                'loctype': slice(107, 108),
                'evtype': slice(109, 111),
            },
            'arrival': {
                'amp': slice(94, 104),
            },
        }
        warnings.simplefilter("ignore", UserWarning)
        catalog = _read_gse2(filename, fields=fields,
                             event_point_separator=True)
        warnings.filters.pop(0)
        self.assertEqual(len(catalog), 2)
        event = catalog[0]
        self.assertEqual(len(event.origins), 1)
        origin = event.origins[0]
        # Test fields with non-standard indexes
        self.assertEqual(origin.creation_info.author, 'bulletin_test')
        self.assertEqual(origin.resource_id, 'smi:local/origin/282672')
        self.assertEqual(origin.quality.standard_error, 0.55)
        uncertainty = origin.origin_uncertainty
        self.assertEqual(uncertainty.azimuth_max_horizontal_uncertainty, 27.05)
        self.assertEqual(origin.evaluation_mode, "manual")
        self.assertEqual(origin.method_id, 'smi:local/method/inversion')
        self.assertEqual(event.event_type, 'earthquake')
        self.assertEqual(event.event_type_certainty, "known")
        self.assertEqual(len(event.picks), 9)
        pick = event.picks[0]
        self.assertEqual(pick.phase_hint, 'Pg')
        self.assertEqual(len(event.amplitudes), 4)
        amplitude = event.amplitudes[0]
        self.assertEqual(amplitude.generic_amplitude, 2.9)

    def test_inventory(self):
        filename = os.path.join(self.path, 'gse_2.0_non_standard.txt')
        inventory_filename = os.path.join(self.path, 'inventory.xml')
        inventory = read_inventory(inventory_filename)
        fields = {
            'line_1': {
                'author': slice(105, 113),
                'id': slice(114, 123),
            },
            'line_2': {
                'az': slice(40, 46),
                'antype': slice(105, 106),
                'loctype': slice(107, 108),
                'evtype': slice(109, 111),
            },
            'arrival': {
                'amp': slice(94, 104),
            },
        }
        warnings.simplefilter("ignore", UserWarning)
        catalog = _read_gse2(filename, inventory, fields=fields,
                             event_point_separator=True)
        warnings.filters.pop(0)
        self.assertEqual(len(catalog), 2)
        # Test a station present in the inventory
        event = catalog[0]
        self.assertEqual(len(event.picks), 9)
        pick = event.picks[0]
        waveform = pick.waveform_id
        self.assertEqual(waveform.network_code, 'ZU')
        self.assertEqual(waveform.channel_code, 'SHZ')
        self.assertEqual(waveform.location_code, '1')
        # Test a station not present in the inventory
        pick_2 = event.picks[2]
        waveform_2 = pick_2.waveform_id
        self.assertEqual(waveform_2.network_code, 'XX')
        self.assertEqual(waveform_2.channel_code, None)
        self.assertEqual(waveform_2.location_code, None)

    def test_inventory_with_multiple_channels(self):
        filename = os.path.join(self.path, 'gse_2.0_non_standard.txt')
        inventory_filename = os.path.join(self.path,
                                          'inventory_multiple_channels.xml')
        inventory = read_inventory(inventory_filename)
        fields = {
            'line_1': {
                'author': slice(105, 113),
                'id': slice(114, 123),
            },
            'line_2': {
                'az': slice(40, 46),
                'antype': slice(105, 106),
                'loctype': slice(107, 108),
                'evtype': slice(109, 111),
            },
            'arrival': {
                'amp': slice(94, 104),
            },
        }
        warnings.simplefilter("ignore", UserWarning)
        catalog = _read_gse2(filename, inventory, fields=fields,
                             event_point_separator=True)
        warnings.filters.pop(0)
        self.assertEqual(len(catalog), 2)
        # Test a station present in the inventory
        event = catalog[0]
        self.assertEqual(len(event.picks), 9)
        pick = event.picks[0]
        waveform = pick.waveform_id
        self.assertEqual(waveform.network_code, 'ZU')
        self.assertEqual(waveform.channel_code, 'HHZ')
        self.assertEqual(waveform.location_code, '1')
        # Test a station with several channels
        pick_2 = event.picks[2]
        waveform_2 = pick_2.waveform_id
        self.assertEqual(waveform_2.network_code, 'ZU')
        self.assertEqual(waveform_2.channel_code, None)
        self.assertEqual(waveform_2.location_code, None)
        # Test a station not present in the inventory
        pick_3 = event.picks[3]
        waveform_3 = pick_3.waveform_id
        self.assertEqual(waveform_3.network_code, 'XX')
        self.assertEqual(waveform_3.channel_code, None)
        self.assertEqual(waveform_3.location_code, None)

    def test_several_begin(self):
        """
        Test with several events.
        """
        filename = os.path.join(self.path, 'gse_2.0_2_begins.txt')
        catalog = _read_gse2(filename)
        self.assertEqual(len(catalog), 2)
        # Test firt event
        event_1 = catalog[0]
        self.assertEqual(event_1.resource_id, 'smi:local/event/280435')
        self.assertEqual(len(event_1.event_descriptions), 1)
        self.assertEqual(
            event_1.event_descriptions[0].text,
            'GREECE-ALBANIA BORDER REGION')
        self.assertEqual(len(event_1.comments), 1)
        self.assertEqual(len(event_1.picks), 9)
        self.assertEqual(len(event_1.amplitudes), 6)
        self.assertEqual(len(event_1.origins), 1)
        self.assertEqual(len(event_1.magnitudes), 2)
        self.assertEqual(len(event_1.station_magnitudes), 4)
        # Test second event
        event_2 = catalog[1]
        self.assertEqual(event_2.resource_id, 'smi:local/event/280436')
        self.assertEqual(len(event_2.event_descriptions), 1)
        self.assertEqual(
            event_2.event_descriptions[0].text,
            'VANCOUVER ISLAND REGION')
        self.assertEqual(len(event_2.comments), 1)
        self.assertEqual(len(event_2.picks), 7)
        self.assertEqual(len(event_2.amplitudes), 5)
        self.assertEqual(len(event_2.origins), 1)
        self.assertEqual(len(event_2.magnitudes), 1)
        self.assertEqual(len(event_2.station_magnitudes), 2)

    def test_read_events(self):
        """
        Tests reading a GSE2.0 document via read_events.
        """
        filename = os.path.join(self.path, 'gse_2.0_2_events.txt')
        catalog = read_events(filename)
        self.assertEqual(len(catalog), 2)

    def test_incomplete_file(self):
        filename = os.path.join(self.path, 'gse_2.0_incomplete.txt')
        self.assertRaises(GSE2BulletinSyntaxError, _read_gse2, filename)


def suite():
    return unittest.makeSuite(BulletinTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA @UnusedWildImport

import os
import unittest

from obspy import read_events, Catalog, UTCDateTime
from obspy.core.event import Event
from obspy.io.focmec.core import _is_focmec, _read_focmec


lst_file_first_comment = '\n'.join((
    "",
    "     Dip,Strike,Rake     76.43    59.08   -64.23",
    "     Dip,Strike,Rake     28.90   174.99  -150.97   Auxiliary Plane",
    "     Lower Hem. Trend, Plunge of A,N     84.99    61.10   329.08    "
    "13.57",
    "     Lower Hem. Trend, Plunge of P,T    358.82    51.71   128.90    "
    "26.95",
    "     B trend, B plunge, Angle:  232.62  25.00 105.00",
    "",
    "          Log10(Ratio)                              Ratio     S Polarity",
    "     Observed  Calculated    Difference  Station     Type     Obs.  "
    "Calc. Flag",
    "      0.8847      0.8950      -0.0103      BLA        SH       R      "
    "R       ",
    "      1.1785      1.0810       0.0975      COR        SH       R      "
    "R       ",
    "      0.6013      0.5442       0.0571      HRV        SH       R      "
    "R       ",
    "      0.3287      0.3666      -0.0379      KEV        SH       L      "
    "L       ",
    "      0.8291      0.9341      -0.1050      KIP        SH       R      "
    "R       ",
    "      0.8033      0.7815       0.0218      KIP        SV       B      "
    "B       ",
    "      1.0783      1.1857      -0.1074      PAS        SH       R      "
    "R       ",
    "      0.2576      0.2271       0.0305      TOL        SH       L      "
    "L       ",
    "     -0.2762     -0.4076       0.1314      TOL        SS       F      "
    "F    NUM",
    "     -0.4283     -0.4503       0.0220      HRV        SS       F      "
    "F       ",
    "     -0.0830      0.0713      -0.1543      KEV        SS       B      "
    "B       ",
    "",
    "Total number of ratios used is  11",
    "RMS error for the 11 acceptable solutions is 0.0852",
    "  Highest absolute velue of diff for those solutions is  0.1543"))
out_file_first_comment = (
    "    Dip   Strike   Rake    Pol: P     SV    SH  AccR/TotR  RMS RErr  "
    "AbsMaxDiff\n   76.43   59.08  -64.23       0.00  0.00  0.00    11/11    "
    "0.0852    0.1543")
creation_time = UTCDateTime(2017, 9, 8, 14, 54, 58)


class FOCMECTestCase(unittest.TestCase):
    """
    Test everything related to reading FOCMEC files
    """
    def setUp(self):
        # Directory where the test files are located
        self.path = os.path.dirname(__file__)
        self.datapath = os.path.join(self.path, 'data')
        self.lst_file = os.path.join(self.datapath, 'focmec_8sta.lst')
        self.out_file = os.path.join(self.datapath, 'focmec_8sta.out')
        with open(self.out_file, 'rb') as fh:
            header = []
            for i in range(15):
                header.append(fh.readline().decode('ASCII'))
            self.out_file_header = ''.join(header).rstrip()
        with open(self.lst_file, 'rb') as fh:
            header = []
            for i in range(48):
                header.append(fh.readline().decode('ASCII'))
            self.lst_file_header = ''.join(header).rstrip()

    def _assert_cat_common_parts(self, cat):
        self.assertTrue(isinstance(cat, Catalog))
        self.assertEqual(len(cat), 1)
        event = cat[0]
        self.assertTrue(isinstance(event, Event))
        self.assertEqual(event.creation_info.creation_time,
                         UTCDateTime(2017, 9, 8, 14, 54, 58))
        self.assertEqual(len(event.focal_mechanisms), 4)
        expected_dip_strike_rake = (
            (76.43, 59.08, -64.23),
            (77.05, 54.08, -59.13),
            (77.05, 59.89, -59.13),
            (77.76, 54.50, -54.06))
        for focmec, (dip, strike, rake) in zip(
                event.focal_mechanisms, expected_dip_strike_rake):
            plane1 = focmec.nodal_planes.nodal_plane_1
            self.assertEqual(plane1.strike, strike)
            self.assertEqual(plane1.dip, dip)
            self.assertEqual(plane1.rake, rake)
            self.assertEqual(focmec.nodal_planes.preferred_plane, 1)
        for focmec in cat[0].focal_mechanisms:
            self.assertEqual(focmec.station_polarity_count, 23)
            # check creation time
            self.assertEqual(focmec.creation_info.creation_time, creation_time)
            # check creation info version
            self.assertEqual(focmec.creation_info.version, 'FOCMEC')

    def _assert_cat_out(self, cat):
        self._assert_cat_common_parts(cat)
        self.assertEqual(cat[0].comments[0].text, self.out_file_header)
        self.assertEqual(cat[0].focal_mechanisms[0].comments[0].text,
                         out_file_first_comment)
        for focmec in cat[0].focal_mechanisms:
            # misfit should be None, because the file specifies that polarity
            # errors are weighted and in the out file format we can't know how
            # many individual errors there are
            self.assertEqual(focmec.misfit, None)
            # we can't tell the gap from the out format
            self.assertEqual(focmec.azimuthal_gap, None)

    def _assert_cat_lst(self, cat):
        self._assert_cat_common_parts(cat)
        self.assertEqual(cat[0].comments[0].text, self.lst_file_header)
        self.assertEqual(cat[0].focal_mechanisms[0].comments[0].text,
                         lst_file_first_comment)
        for focmec in cat[0].focal_mechanisms:
            # misfit should be 0.0, because in the lst file we can count the
            # number of individual errors (and there's no polarity errors)
            self.assertEqual(focmec.misfit, 0.0)
            # we can't tell the gap from the out format
            self.assertEqual(focmec.azimuthal_gap, 236.7)

    def test_is_focmec(self):
        for file_ in (self.lst_file, self.out_file):
            self.assertTrue(_is_focmec(file_))

    def test_read_focmec_out(self):
        cat = _read_focmec(self.out_file)
        self._assert_cat_out(cat)

    def test_read_focmec_out_open_file(self):
        for mode in ('rb', 'rt'):
            with open(self.out_file, mode) as fh:
                cat = _read_focmec(fh)
            self._assert_cat_out(cat)

    def test_read_focmec_lst(self):
        cat = _read_focmec(self.lst_file)
        self._assert_cat_lst(cat)

    def test_read_focmec_lst_open_file(self):
        for mode in ('rb', 'rt'):
            with open(self.lst_file, mode) as fh:
                cat = _read_focmec(fh)
            self._assert_cat_lst(cat)

    def test_read_focmec_out_through_plugin(self):
        cat = read_events(self.out_file)
        self._assert_cat_out(cat)

    def test_read_focmec_lst_through_plugin(self):
        cat = read_events(self.lst_file)
        self._assert_cat_lst(cat)

    def test_read_focmec_lst_other_flavors(self):
        """
        This tests some additional files. lst files have a pretty free format
        unfortunately. It depends on focmec program version and on input data /
        program options what lines get printed and also how specific lines we
        look for look like..

        For these files just test a few things, in general reading should be
        covered by basic tests above
        """
        # 1: focmec_qedUWne.lst
        cat = read_events(os.path.join(self.datapath, 'focmec_qedUWne.lst'))
        self.assertEqual(len(cat), 1)
        focmecs = cat[0].focal_mechanisms
        self.assertEqual(len(focmecs), 5)
        focmec = focmecs[0]
        plane = focmec.nodal_planes.nodal_plane_1
        self.assertEqual(plane.strike, 308.43)
        self.assertEqual(plane.dip, 58.68)
        self.assertEqual(plane.rake, 16.48)
        for focmec in focmecs:
            self.assertEqual(focmec.azimuthal_gap, 79.5)
        for focmec in focmecs[:-1]:
            self.assertEqual(focmec.station_polarity_count, 190)
            self.assertEqual(focmec.misfit, 21.0 / 190)
        # last solution has one less polarity error
        self.assertEqual(focmecs[-1].station_polarity_count, 190)
        self.assertEqual(focmecs[-1].misfit, 20.0 / 190)

        # 2: focmec_all.lst
        cat = read_events(os.path.join(self.datapath, 'focmec_all.lst'))
        self.assertEqual(len(cat), 1)
        focmecs = cat[0].focal_mechanisms
        self.assertEqual(len(focmecs), 3)
        focmec = focmecs[0]
        plane = focmec.nodal_planes.nodal_plane_1
        self.assertEqual(plane.strike, 66.98)
        self.assertEqual(plane.dip, 80.61)
        self.assertEqual(plane.rake, -69.72)
        for focmec, errors in zip(focmecs, (25, 29, 26)):
            self.assertEqual(focmec.azimuthal_gap, 44.7)
            self.assertEqual(focmec.station_polarity_count, 212)
            self.assertEqual(focmec.misfit, float(errors) / 212)

        # 3: focmec_8sta-noratios.lst
        cat = read_events(os.path.join(self.datapath,
                                       'focmec_8sta-noratios.lst'))
        self.assertEqual(len(cat), 1)
        focmecs = cat[0].focal_mechanisms
        self.assertEqual(len(focmecs), 184)
        focmec = focmecs[2]
        plane = focmec.nodal_planes.nodal_plane_1
        self.assertEqual(plane.strike, 255.00)
        self.assertEqual(plane.dip, 10.0)
        self.assertEqual(plane.rake, -90.00)
        for focmec in focmecs:
            self.assertEqual(focmec.azimuthal_gap, 236.7)
            self.assertEqual(focmec.station_polarity_count, 23)
            self.assertEqual(focmec.misfit, 0.0)


def suite():
    return unittest.makeSuite(FOCMECTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')
